#!/s/perl-5.004/bin/perl -w

# $Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/ag_perl_sql_lexer.pm,v 1.1.1.1 1998/05/05 23:52:24 glew Exp $

package ag_perl_sql_lexer;

# Deduce location of libraries distributed with package
use FindBin;
BEGIN
{
    my $install_root = $FindBin::RealBin;
    $install_root =~ s:(.*/perl-sql)/.*:$1:;
    unshift (@INC, "$install_root/external-dependencies/IO-stringy-1.104/lib");
}


use strict;



#use Fstream;
use IO::File;
use IO::Scalar;
use Data::Dumper;
use Perl_SQL::dumb_line_re_lexer;
use Perl_SQL::ag_perl_sql_parser;
use Carp;


#############################################################################################
# Perl/SQL lexer patterns



my $dec_num = '(?:[0-9]+|[0-9]+\.|[0-9]*\.[0-9]+)';

sub keyword_case_re {
    # Return a string that is a regexp to match keywords
    # subject to my standard case rules:
    # either all uppercase, or all lowercase

    # TBD: should I allow initial letter keywords?
    # Or, should I match mixed case keywords, but report an error for them?
    # (as opposed to allowing mixed case keywords to be identifiers - which is bad?)
    # In the latter case, in general, the pattern that the lexer matches
    # is not the single determinant - there should be a layer of indirection
    # to enforce further rules.  For now, punt.


    my($arg) = @_;

    if( ($arg =~ m/.*[a-z].*/) && ($arg =~ m/.*[A-Z].*/) ) {
	confess "lexer pattern setup error: keywords should not be of mixed case: $arg\n"
	    . "(TBD: although I do not yet error if they are in the input.)\n";
    }


    my $c;
    my $case_insensitive_pattern = '';
    foreach $c ( split(//,$arg) ) {
	if( $c =~ m/[a-zA-Z]/ ) {
	    my $uc = $c;
	    $uc =~ tr/a-z/A-Z/;
	    my $lc = $c;
	    $lc =~ tr/A-Z/a-z/;
	    $case_insensitive_pattern .= "[$lc$uc]";
	}
	elsif( $c =~ m/\[/ ) {
	    carp "lexer pattern setup error: keywords should not contain \[\n";
	}
	else {
	    $case_insensitive_pattern .= $c;
	}
    }

    return $case_insensitive_pattern;
}


$ag_perl_sql_lexer::pattern_tuple_list = 
[
 { token => 'whitespace', regexp => '(?:\s|\n)+', action => 'IGNORE' },
 { token => 'empty_line', regexp => '', action => 'IGNORE' },

 # Single special character tokens (note that character may also appear in multiple character versions)
 { token => 'self', regexp => '[+]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[-]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[*]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[/]', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::EQ, regexp => '[=]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[!]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[~]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[%]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '\^', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[&]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[|]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[(]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[)]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[{]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[}]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '\[', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '\]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[:]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[;]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[,]', action => 'RETURN' }, # Perl
 { token => 'self', regexp => '[.]', action => 'RETURN' }, # Perl

 # Multiple special character tokens, typically operators
 { token => $ag_perl_sql_parser::ANDAND, regexp => '\&\&', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::POW, regexp => '\*\*', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::OROR, regexp => '\|\|', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::EQEQ, regexp => '==', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::MATCH, regexp => '=~', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::NE, regexp => '!=', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::NE, regexp => '<>', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::GT, regexp => '>', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::LT, regexp => '<', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::GE, regexp => '>=', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::LE, regexp => '<=', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::LSHFT, regexp => '<<', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::RSHFT, regexp => '>>', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::ARSHFT, regexp => '>>>', action => 'RETURN' }, # Java
 { token => $ag_perl_sql_parser::ANDk, regexp => keyword_case_re('and'), action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::ORk, regexp => '[Oo][Rr]', action => 'RETURN' }, # SQL
 { token => $ag_perl_sql_parser::ANDTHENk, regexp => keyword_case_re('andthen'), action => 'RETURN' }, # AG
 { token => $ag_perl_sql_parser::ORELSEk, regexp => keyword_case_re('orelse'), action => 'RETURN' }, # AG
 { token => $ag_perl_sql_parser::EQk, regexp => keyword_case_re('eq'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::NEk, regexp => keyword_case_re('ne'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::GTk, regexp => keyword_case_re('gt'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::GEk, regexp => keyword_case_re('ge'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::LTk, regexp => keyword_case_re('lt'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::LEk, regexp => keyword_case_re('le'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::DIV, regexp => keyword_case_re('div'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::MOD, regexp => keyword_case_re('mod'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::REM, regexp => keyword_case_re('rem'), action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::XOR, regexp => keyword_case_re('xor'), action => 'RETURN' }, # Perl

 { token => $ag_perl_sql_parser::SET_COLON_EQUALS, regexp => ':=', action => 'RETURN' }, # Perl
 { token => $ag_perl_sql_parser::SET_RIGHT_DOUBLE_ARROW, regexp => '=>', action => 'RETURN' }, # Perl


 # Keywords
    { token => $ag_perl_sql_parser::ALL, regexp => keyword_case_re("ALL"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ALTER, regexp => keyword_case_re("ALTER"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::AS, regexp => keyword_case_re("AS"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ASCENDING, regexp => keyword_case_re("ASC"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ASCENDING, regexp => keyword_case_re("ASCENDING"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::BETWEEN, regexp => keyword_case_re("BETWEEN"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::BY, regexp => keyword_case_re("BY"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::CASE, regexp => keyword_case_re("CASE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::COLUMN, regexp => keyword_case_re("COLUMN"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::COUNT, regexp => keyword_case_re("COUNT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::DELETE, regexp => keyword_case_re("DELETE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::DESCENDING, regexp => keyword_case_re("DESC"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::DESCENDING, regexp => keyword_case_re("DESCENDING"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::DISTINCT, regexp => keyword_case_re("DISTINCT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::DROP, regexp => keyword_case_re("DROP"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ELSE, regexp => keyword_case_re("ELSE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::END, regexp => keyword_case_re("END"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::EXISTS, regexp => keyword_case_re("EXISTS"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::FROM, regexp => keyword_case_re("FROM"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::FULL, regexp => keyword_case_re("FULL"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::GOTO, regexp => keyword_case_re("GOTO"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::GROUP, regexp => keyword_case_re("GROUP"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::HAVING, regexp => keyword_case_re("HAVING"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::IN, regexp => keyword_case_re("IN"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::INSERT, regexp => keyword_case_re("INSERT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::INTO, regexp => keyword_case_re("INTO"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::IS, regexp => keyword_case_re("IS"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::NOTk, regexp => keyword_case_re("NOT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::NULL, regexp => keyword_case_re("NULL"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::OF, regexp => keyword_case_re("OF"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ON, regexp => keyword_case_re("ON"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::ORDER, regexp => keyword_case_re("ORDER"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::SELECT, regexp => keyword_case_re("SELECT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::SET, regexp => keyword_case_re("SET"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::TABLE, regexp => keyword_case_re("TABLE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::THEN, regexp => keyword_case_re("THEN"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::TO, regexp => keyword_case_re("TO"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::UNION, regexp => keyword_case_re("UNION"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::UNIQUE, regexp => keyword_case_re("UNIQUE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::UPDATE, regexp => keyword_case_re("UPDATE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::USING, regexp => keyword_case_re("USING"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::VALUES, regexp => keyword_case_re("VALUES"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::WHEN, regexp => keyword_case_re("WHEN"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::WHERE, regexp => keyword_case_re("WHERE"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::WITH, regexp => keyword_case_re("WITH"), action => 'RETURN' },

    { token => $ag_perl_sql_parser::FULL, regexp => keyword_case_re("FULL"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::OUTER, regexp => keyword_case_re("OUTER"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::INNER, regexp => keyword_case_re("INNER"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::LEFT, regexp => keyword_case_re("LEFT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::RIGHT, regexp => keyword_case_re("RIGHT"), action => 'RETURN' },
    { token => $ag_perl_sql_parser::JOIN , regexp => keyword_case_re("JOIN "), action => 'RETURN' },

    { token => $ag_perl_sql_parser::UNDEF, regexp => "undef", action => 'RETURN' }, # AG extension - Perl undef


 # Quoted strings
 { token => $ag_perl_sql_parser::STRING, regexp => "' (?:[^']|'')*'", action => 'RETURN' },
 # Errors: strings past end of line
 # (use string concatenation for multiline strings)
 #broken: { token => 'string_past_eoln', regexp => "' [^'\]* (?:\' [^']*)*", action => 'ERROR:string-eoln' },
 #broken: { token => 'string_past_eoln', regexp => '" [^"\\]* (?:\" [^"]*)*', action => 'ERROR:string-eoln' },

 # Numbers
 { token => $ag_perl_sql_parser::NUMBER, subtype => 'octal integer',
   regexp => '\b0[0-7]+', action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER, subtype=>'decimal integer', regexp => '\b' . $dec_num . '\b', action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER,  subtype=>'decimal float', 
   regexp => '\b' . $dec_num . '[Ee](?:\+|\-|)[0-9]+\b', action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER, subtype => 'hexadecimal integer',
   regexp => '\b0x[0-9a-fA-F]+', action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER, subtype => 'hexadecimal integer',
   regexp => "X'[0-9a-fA-F]+'", action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER, subtype => 'binary integer',
   regexp => "B'[01]+'", action => 'RETURN' },
 { token => $ag_perl_sql_parser::NUMBER, subtype => 'arbitrary base integer',
   regexp => '[0-9]+\#[0-9a-zA-Z]+', action => 'RETURN' },

 # Identifiers, including delimited identifiers (strings)
 { token => $ag_perl_sql_parser::DELIMITED_IDENTIFIER, regexp => '"(?:[^"]|"")*"', action => 'RETURN' },
 { token => $ag_perl_sql_parser::IDENTIFIER, regexp => '\b[A-Za-z_][A-Za-z_0-9]*\b', action => 'RETURN' },

 ];






#############################################################################################
# interactive tests - command line control


my $verbose=0;


sub test_driver {

    my @ARGV2 = @ARGV;
    my $arg;

    while( $arg = shift @ARGV2 ) {
	if( 0 ) {
	}
	elsif( $arg eq "-verbose" ) {
	    $verbose = 1;
	}
	elsif( $arg eq "-verbose=0" ) {
	    $verbose = 0;
	}
	elsif( $arg eq "-verbose=1" ) {
	    $verbose = 1;
	}
	else {
	    die "undefined command line argument <$arg>\n";
	}

    }
    built_in_tests();
}

#############################################################################################
# built in tests




sub built_in_tests {


    my @tests = (
		 {    string => "xor",
		      expect => [ 
				 { token => $ag_perl_sql_parser::XOR, string => "xor", },
				 ],
		      },
		 {    string => "0",
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => "0", },
				 ],
		      },
		 {    string => "1",
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => "1", },
				 ],
		      },
		 {    string => "alter table t1 drop column c1",
		      expect => [ 
				 { token => $ag_perl_sql_parser::ALTER, string => "alter", },
				 { token => $ag_perl_sql_parser::TABLE, string => "table", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "t1", },
				 { token => $ag_perl_sql_parser::DROP, string => "drop", },
				 { token => $ag_perl_sql_parser::COLUMN, string => "column", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "c1", },
				 ],
		      },
		 {    string => "a:=b, c=>d",
		      expect => [ 
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "a", },
				 { token => $ag_perl_sql_parser::SET_COLON_EQUALS, string => ":=", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "b", },
				 { token => 'self', string => ",", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "c", },
				 { token => $ag_perl_sql_parser::SET_RIGHT_DOUBLE_ARROW, string => "=>", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "d", },
				 ],
		      },
		 #BUG: I cannot seem to handle both '' and \' inside a single quoted string
		 #{    string => "'string1\\'string2'",
		 #     expect => [ 
		 #debug		 { token => $ag_perl_sql_parser::STRING, string => "'string1\\'string2'", },
		 #debug		 ],
		 #debug     },
		 {    string => "'string1''string2'",
		      expect => [ 
				 { token => $ag_perl_sql_parser::STRING, string => "'string1''string2'", },
				 ],
		      },
		 {    string => "'string1' 'string2'",
		      expect => [ 
				 { token => $ag_perl_sql_parser::STRING, string => "'string1'", },
				 { token => $ag_perl_sql_parser::STRING, string => "'string2'", },
				 ],
		      },
		 {    string => "WHERE actor_last_name = 'Moore' AND actor_first_name = 'Demi'",
		      expect => [ 
				 { token => $ag_perl_sql_parser::WHERE, string => "WHERE", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "actor_last_name", },
				 { token => $ag_perl_sql_parser::EQ, string => "=", },
				 { token => $ag_perl_sql_parser::STRING, string => "'Moore'", },
				 { token => $ag_perl_sql_parser::ANDk, string => "AND", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => "actor_first_name", },
				 { token => $ag_perl_sql_parser::EQ, string => "=", },
				 { token => $ag_perl_sql_parser::STRING, string => "'Demi'", },
				 ],
		      },
		 {    string => "DELETE",
		      expect => [ 
				 { token => $ag_perl_sql_parser::DELETE, string => "DELETE", },
				 ],
		      },
		 {    string => "X'DeadBeef'",
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => "X'DeadBeef'", },
				 ],
		      },
		 {    string => "B'01001'",
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => "B'01001'", },
				 ],
		      },
		 {    string => '111',
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => '111', },
				 ],
		      },
		 {    string => 'identiii',
		      expect => [ 
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'identiii', },
				 ],
		      },
		 {    string => 'identIII',
		      expect => [ 
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'identIII', },
				 ],
		      },
		 {    string => 'or',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ORk, string => 'or', },
				 ],
		      },
		 {    string => 'OR',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ORk, string => 'OR', },
				 ],
		      },
		 {    string => 'and',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ANDk, string => 'and', },
				 ]
		      },
		 {    string => 'or',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ORk, string => 'or', },
				 ],
		      },
		 {    string => 'andthen',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ANDTHENk, string => 'andthen', },
				 ],
		      },
		 {    string => 'orelse',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ORELSEk, string => 'orelse', },
				 ],
		      },
		 {    string => 'eq',
		      expect => [ 
				 { token => $ag_perl_sql_parser::EQk, string => 'eq', },
				 ],
		      },
		 {    string => 'ne',
		      expect => [ 
				 { token => $ag_perl_sql_parser::NEk, string => 'ne', },
				 ],
		      },
		 {    string => 'gt',
		      expect => [ 
				 { token => $ag_perl_sql_parser::GTk, string => 'gt', },
				 ],
		      },
		 {    string => 'ge',
		      expect => [ 
				 { token => $ag_perl_sql_parser::GEk, string => 'ge', },
				 ],
		      },
		 {    string => 'lt',
		      expect => [ 
				 { token => $ag_perl_sql_parser::LTk, string => 'lt', },
				 ],
		      },
		 {    string => 'le',
		      expect => [ 
				 { token => $ag_perl_sql_parser::LEk, string => 'le', },
				 ],
		      },
		 {    string => 'and',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ANDk, string => 'and', },
				 ],
		      },
		 {    string => 'and',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ANDk, string => 'and', },
				 ],
		      },
		 {    string => '1 2 3',
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => '1', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '2', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '3', },
				 ],
		      },
		 {    string => '-+*/=!%^&()|{}[]:;,.<>&&||**==!=<>>=<=<<>> >>> >>',
		      expect => [ 
				 { token => 'self', string => '-', },
				 { token => 'self', string => '+', },
				 { token => 'self', string => '*', },
				 { token => 'self', string => '/', },
				 { token => $ag_perl_sql_parser::EQ, string => '=', },
				 { token => 'self', string => '!', },
				 { token => 'self', string => '%', },
				 { token => 'self', string => '^', },
				 { token => 'self', string => '&', },
				 { token => 'self', string => '(', },
				 { token => 'self', string => ')', },
				 { token => 'self', string => '|', },
				 { token => 'self', string => '{', },
				 { token => 'self', string => '}', },
				 { token => 'self', string => '[', },
				 { token => 'self', string => ']', },
				 { token => 'self', string => ':', },
				 { token => 'self', string => ';', },
				 { token => 'self', string => ',', },
				 { token => 'self', string => '.', },
				 { token => $ag_perl_sql_parser::NE, string => '<>', },
				 { token => $ag_perl_sql_parser::ANDAND, string => '&&', },
				 { token => $ag_perl_sql_parser::OROR, string => '||', },
				 { token => $ag_perl_sql_parser::POW, string => '**', },
				 { token => $ag_perl_sql_parser::EQEQ, string => '==', },
				 { token => $ag_perl_sql_parser::NE, string => '!=', },
				 { token => $ag_perl_sql_parser::NE, string => '<>', },
				 { token => $ag_perl_sql_parser::GE, string => '>=', },
				 { token => $ag_perl_sql_parser::LE, string => '<=', },
				 { token => $ag_perl_sql_parser::LSHFT, string => '<<', },
				 { token => $ag_perl_sql_parser::RSHFT, string => '>>', },
				 { token => $ag_perl_sql_parser::ARSHFT, string => '>>>', },
				 { token => $ag_perl_sql_parser::RSHFT, string => '>>', },
				 ]
				 },
		 {    string => '126 && 0x6Ab6778 ** 077 || 0 <> 2#00101',
		      expect => [ 
				 { token => $ag_perl_sql_parser::NUMBER, string => '126', },
				 { token => $ag_perl_sql_parser::ANDAND, string => '&&', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '0x6Ab6778', },
				 { token => $ag_perl_sql_parser::POW, string => '**', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '077', },
				 { token => $ag_perl_sql_parser::OROR, string => '||', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '0', },
				 { token => $ag_perl_sql_parser::NE, string => '<>', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '2#00101', },
				 ]
				 },
		 {    string => '"aa"bb',
		      expect => [ 
				 { token => $ag_perl_sql_parser::DELIMITED_IDENTIFIER, string => '"aa"', },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'bb', },
				 ]
				 },
		 #BUG: I cannot seem to handle both '' and \' inside a single quoted string
		 #debug{    string => '"aa\"bb"cc',
		 #debug     expect => [ 
		 #debug		 { token => $ag_perl_sql_parser::DELIMITED_IDENTIFIER, string => '"aa\"bb"', },
		 #debug		 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'cc', },
		 #debug		 ]
		 #debug		 },
		 #BUG: I cannot seem to handle both '' and \' inside a single quoted string
		 #debug{    string => "'aa\'bb'cc",
		 #debug     expect => [ 
		 #debug		 { token => $ag_perl_sql_parser::STRING, string => "'aa\'bb'", },
		 #debug		 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'cc', },
		 #debug		 ]
		 #debug		 },
		 {    string => "'aa''bb'cc",
		      expect => [ 
				 { token => $ag_perl_sql_parser::STRING, string => "'aa''bb'", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'cc', },
				 ]
				 },
		 {    string => '1+2<<3*675657-ifee+0xf00a',
		      expect => [
				 { token => $ag_perl_sql_parser::NUMBER, string => '1', },
				 { token => 'self', string => '+', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '2', },
				 { token => $ag_perl_sql_parser::LSHFT, string => '<<', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '3', },
				 { token => 'self', string => '*', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '675657', },
				 { token => 'self', string => '-', },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'ifee', },
				 { token => 'self', string => '+', },
				 { token => $ag_perl_sql_parser::NUMBER, string => '0xf00a', },
				 ]
				 },
		 {    string => 'aa bb',
		      expect => [
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'aa', },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'bb', },
				 ]
				 },
		 {    string => "'aa'bb",
		      expect => [
				 { token => $ag_perl_sql_parser::STRING, string => "'aa'", },
				 { token => $ag_perl_sql_parser::IDENTIFIER, string => 'bb', },
				 ]
				 },
		 {    string => 'and or andthen orelse eq ne gt ge lt le',
		      expect => [ 
				 { token => $ag_perl_sql_parser::ANDk, string => 'and', },
				 { token => $ag_perl_sql_parser::ORk, string => 'or', },
				 { token => $ag_perl_sql_parser::ANDTHENk, string => 'andthen', },
				 { token => $ag_perl_sql_parser::ORELSEk, string => 'orelse', },
				 { token => $ag_perl_sql_parser::EQk, string => 'eq', },
				 { token => $ag_perl_sql_parser::NEk, string => 'ne', },
				 { token => $ag_perl_sql_parser::GTk, string => 'gt', },
				 { token => $ag_perl_sql_parser::GEk, string => 'ge', },
				 { token => $ag_perl_sql_parser::LTk, string => 'lt', },
				 { token => $ag_perl_sql_parser::LEk, string => 'le', },
				 ]
				 },
		 )
    ;




    my $test_pattern;
    foreach $test_pattern ( @tests ) {

	my $s = IO::Scalar->new( \$$test_pattern{string} );
	my $lex_stream = dumb_line_re_lexer->new( pattern_table=>$ag_perl_sql_lexer::pattern_tuple_list, stream=>$s );

	my @expect_list = @{$test_pattern->{expect}};


	my $token;
	while( $token = $lex_stream->get_token ) {
	    my $e = shift @expect_list;

	    if( $token && $e && $token->{token} eq $e->{token} && $token->{string} eq $e->{string} ) {
		# subtest passed
	    }
	    else {
		# subtest failed
		die "****************************\nSubtest failed:\n"
		    . Data::Dumper->Dump([$token], [qw(parsed_token)])
		    . Data::Dumper->Dump([$e], [qw(expected_token)])
		    . Data::Dumper->Dump([$test_pattern], [qw($test_pattern)])
		    ;
	    }
	}
	if( @expect_list ) {
	    die "****************************\nTest failed: early termination\n"
		. Data::Dumper->Dump([$test_pattern], [qw($test_pattern)]);
	}
	else {
	    if( $verbose == 0 ) {
		print "Test passed: $test_pattern->{string}\n";
	    }
	    else {
		print "Test passed:\n"
		    . Data::Dumper->Dump([$test_pattern], [qw($test_pattern)])
		    . "\n";
	    }
	}

    }
}


"Packages return non-zero";
