%{
package ag_perl_sql_parser;
use Perl_SQL::dumb_line_re_lexer;
use Data::Dumper;
use Carp;
use Perl_SQL::eval_functions;
%}

#

%token NUMBER
%token STRING
%token IDENTIFIER
%token DELIMITED_IDENTIFIER

# Standard SQL 2 keywords

%token ALTER
%token ALL      
%token AS
%token ASCENDING        # Glew extension to ASC
%token BY
%token BETWEEN
%token CASE
%token COLUMN
%token COUNT
%token DESCENDING       # Glew extension to DESC
%token DISTINCT
%token DROP
%token END
%token ELSE
%token EXISTS
%token FROM
%token GOTO
%token GROUP
%token HAVING
%token IN
%token INSERT
%token INTO
%token IS
%token NULL
%token OF
%token ON
%token ORDER
%token SELECT
%token SET
%token TABLE
%token THEN
%token TO
%token UNION
%token UNIQUE
%token UPDATE
%token USING
%token VALUES
%token WHEN
%token WHERE
%token WITH

# Standard extended SQL keywords - mainly for outer joins

%token OUTER
%token INNER
%token LEFT
%token RIGHT
%token JOIN
%token FULL

# AG extensions

%token UNDEF
%token XOR
%token SET_COLON_EQUALS
%token SET_RIGHT_DOUBLE_ARROW


# Operators - both standard SQL< and AG extensions (Perl++)
%left  UNION
%nonassoc SELECT UPDATE DELETE
%nonassoc FROM WHERE INTO
%left IN
%left FULL OUTER INNER LEFT RIGHT JOIN
%nonassoc ON
%nonassoc AS
%left  ORk  ORELSEk XOR /* or */
%left  ANDk ANDTHENk /* and */
%right NOTk         /* not */
%left  OROR         /* || */
%left  ANDAND       /* && */
%right BETWEEN
%left  '|' '^'      /* bit */
%left  '&'          /* bit */
%nonassoc EQEQ LT GT EQ NE GE LE EQk NEk GTk LTk LEk GEk  /* == = eq etc. */
%left LSHFT RSHFT ARSHFT /* << >> >>> */
%left '+' '-'       /* + - */
%left '*' '/' DIV '%' MOD REM /* * / */
%left  MATCH        /* =~ */
%right '!' '~' UMINUS
%right  POW         /* ** */
%left '.'           /* field separator */
%left '(' '[' '{'



%%

# expr
#   I have been oscillating between being purely expression oriented,
# relying on external code to do type checking,
# and between using YACC to do trivial checking, e.g. that an argumen
# is a table_expr or a scalar_expr.
#   Currently I take the latter course, and disable the top level expr
# to avoid ambiguities in the grammar (reduce/reduce and shift/reduce
# conflicts) -- or, rather, I do not use such the top level expr
# (now accept_expr) fully recursively.

accept_expr:            scalar_expr     { $$ = $1; }
|			table_expr      { $$ = $1; }
;


scalar_expr:    NUMBER          { $$ = add_fields( $1, {  interpret_string => ("number(" . deref($1) . ")"),
							  eval_function => \&eval_functions::eval_NUMBER,
							  eval_name => deref($1), /* TBD: value, for non-Perl syntaxes? */
                                                       } );
                                }
|               STRING          { $$ = add_fields( $1, {  interpret_string => ("string(" . deref($1) . ")"),
							  eval_function => \&eval_functions::eval_STRING,
							  eval_name => deref($1), /* TBD: value, for non-Perl syntaxes? */
                                                       } );
                                }
|               column_ref      { $$ = $1; }
|               infix_op_expr   { $$ = $1; }
|               prefix_op_expr  { $$ = $1; }
|               '(' scalar_expr ')'     { $$ = $2; }
|               function_call   { $$ = $1; }
|               case_expr       { $$ = $1; }
|               funky_syntax_predicates { $$ = $1; }
|		count_expr	{ $$ = $1; }
;


funky_syntax_predicates:
        scalar_expr BETWEEN scalar_expr ANDk scalar_expr
            { $$ = {  interpret_string => ('BETWEEN(value=' . deref($1) . ',low=' . deref($3) . ',up=' . deref($5) . ')'),
                                          eval_function => \&eval_functions::eval_ERROR,
                                               };
                                }
|       scalar_expr NOTk BETWEEN scalar_expr ANDk scalar_expr
            { $$ = {  interpret_string => ('NOT(BETWEEN(value=' . deref($1) . ',low=' . deref($4) . ',up=' . deref($6) . '))'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|       column_ref IS NULL { $$ = {  interpret_string => ('IS_NULL(' . deref($1) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|       column_ref IS NOTk NULL { $$ = {  interpret_string => ('NOT(IS_NULL(' . deref($1) . '))'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


case_expr:  predicate_case_expr { $$ = $1; }
|           value_case_expr { $$ = $1; }
;


predicate_case_expr: CASE case_when_list case_else END { $$ = {  interpret_string => ('CASE_PREDICATE(when=(' . deref($2) . '),else=' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


value_case_expr: CASE scalar_expr case_when_list case_else END
    { $$ = {  interpret_string => ('CASE_VALUE(value=' . deref($2) . ',when=(' . deref($3) . '),else=' . deref($4) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

case_when_list: case_when { $$ = $1; }
|               case_when_list case_when { $$ = {  interpret_string => (deref($1) . ',' . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;
case_when: WHEN scalar_expr THEN scalar_expr { $$ = {  interpret_string => ('(when=' . deref($2) . ',then=' . deref($4) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

case_else: ELSE scalar_expr { $$ = {  interpret_string => (deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;



count_expr:	COUNT '(' '*' ')'		    { $$ = { interpret_string => 'COUNT(*)',
							     eval_name => 'COUNT(*)',
							     eval_function => \&eval_functions::eval_COUNT_STAR,
							     is_aggregate => 1,  # is definitely an aggregate function
							   };
						    }
;


function_call:	
    simple_identifier '(' scalar_expr_commalist ')'
	{
	    my $eval_name = $1->{eval_name} . "(" . $3->{eval_name} . ")";
	    if( eval_functions::is_aggregate_function_name($eval_name) ) {
		$$ = { interpret_string => ('AGGREGATE(func=' . deref($1) . ",args=(" . deref($3) . "))"),
		       eval_function => \&eval_functions::eval_AGGREGATE_FUNCTION,
		       function_name => $1,
		       argument_list => $3,
		       eval_name => $eval_name,
		       is_aggregate => 1,  # TBD: currently all functions are assumed to be aggregates
		       #is_aggregate => -1,  # is possibly an aggregate function
		   }
	    }
	    else {
		$$ = { interpret_string => ('FUNCTION_CALL(func=' . deref($1) . ",args=(" . deref($3) . "))"),
		       eval_function => \&eval_functions::eval_FUNCTION,
		       function_name => $1,
		       argument_list => $3,
		       eval_name => $eval_name,
		   }
	    }
	}
# Special handling for COUNT
# mainly because COUNT(*) syntax was awkward for standard functions
|   COUNT '(' scalar_expr_commalist ')'
	    { $$ = { interpret_string => ('FUNCTION_CALL(func=' . "COUNT" . ",args=(" . deref($3) . "))"),
		     eval_function => \&eval_functions::eval_AGGREGATE_FUNCTION,
		     function_name => "COUNT",
		     argument_list => $3->{list},
		     eval_name => "COUNT" . "(" . $3->{eval_name} . ")",
		     is_aggregate => 1,  # is definitely an aggregate function
		   };
	    }
;


wildcard:   '*'  { $$ = {  interpret_string => "WILDCARD('*')",
                           eval_name => "WILDCARD(\'*\')",
                           eval_function => \&eval_functions::eval_ERROR,
                        };
                 }
            /* TBD: other wildcards, like T1.*, or even *addr* */
;





table_expr: table_identifier { $$ = $1; }
|     join_table_expression { $$ = $1; }
|     complex_table_expr    { $$ = $1; }
|     '(' table_expr ')'      { $$ = $2;
				  $$->{eval_name} = ('(' . $2->{eval_name} . ')');
				}
|     table_expr UNION table_expr       { $$ = {  interpret_string => ("UNION(" . deref($1) . "," . deref($3) . ")"),
                                                  eval_function => \&eval_functions::eval_ERROR,
						  eval_name => ($1->{eval_name} . ' union ' . $3->{eval_name}),
                                               };
}
;



prefix_op_expr:	    '+' scalar_expr %prec UMINUS { $$ = $2;    # unary PLUS is a nop, present for symmetry with unary MINUS
						   $$->{eval_name} = ('+' . $2->{eval_name}),
                                                 }
|                   '-' scalar_expr %prec UMINUS { $$ = {  interpret_string => ("-" . deref($2)),
							   eval_function => \&eval_functions::eval_null_propagating_perl_prefix_operator,
							   perl_operator => '-', 
						           eval_name => ('-' . $2->{eval_name}),
							   operand => $2,
                                                        };
                                                 }
|                   '!'  scalar_expr    { $$ = {  interpret_string => ("NOT_EXCLAM(" . deref($2) .")"),
						  eval_function => \&eval_functions::eval_null_propagating_perl_prefix_operator,
						  perl_operator => '!', 
					          operand => $2,
						  eval_name => ('!' . $2->{eval_name}),
                                                };
                                        }
|                   '~'  scalar_expr    { $$ = {  interpret_string => ("NOT_TILDE(" . deref($2) .")"),
						  eval_function => \&eval_functions::eval_null_propagating_perl_prefix_operator,
						  perl_operator => '~', 
					          operand => $2,
						  eval_name => ('~' . $2->{eval_name}),
                                                };
                                        }
|                   NOTk  scalar_expr   { $$ = {  interpret_string => ("NOT(" . deref($2) .")"),
						  eval_function => \&eval_functions::eval_null_propagating_perl_prefix_operator,
						  perl_operator => 'not', 
					          operand => $2,
						  eval_name => ('not ' . $2->{eval_name}),
                                                };
                                        }
|                   EXISTS  subquery    { $$ = {  interpret_string => ("EXISTS(" . deref($2) .")"),
                                                  eval_function => \&eval_functions::eval_ERROR,
						  eval_name => ('exists ' . $2->{eval_name}),
                                               };
                                                 }
|                   UNIQUE  subquery    { $$ = {  interpret_string => ("UNIQUE(" . deref($2) .")"),
                                                  eval_function => \&eval_functions::eval_ERROR,
						  eval_name => ('unique ' . $2->{eval_name}),
                                               };
                                                 }
# disabled: conflicts with SELECT ALL / SELECT DISTINCT
# => need to fix the grammar
#|                   quantifier  scalar_expr     { $$ = {  interpret_string => (deref($1) . "(" . deref($2) .")"),
#                                                  eval_function => \&eval_functions::eval_ERROR,
#                                               };
#                                                 }
        #disabled because (1) causes conflicts, and (2) not implementing yet
        #disabled TBD: fix
        #disabled|           quantifier  table_expr     { $$ = {  interpret_string => (deref($1) . "(" . deref($2) .")"),
        #disabled                                                 eval_function => \&eval_functions::eval_ERROR,
        #disabled                                              };
        #disabled                                       }
;

infix_op_expr:
                scalar_expr ORk scalar_expr     { $$ = {  interpret_string => ("ORk(" . deref($1) . "," . deref($3) .")"),
							  eval_function => \&eval_functions::eval_standard_perl_infix_operator,
							  eval_name => ($1->{eval_name} . ' or ' . $3->{eval_name}),
							  perl_operator => 'or', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr XOR scalar_expr     { $$ = {  interpret_string => ("XOR(" . deref($1) . "," . deref($3) .")"),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  eval_name => ($1->{eval_name} . ' xor ' . $3->{eval_name}),
							  perl_operator => '^', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr ANDk scalar_expr    { $$ = {  interpret_string => ("ANDk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' and ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_standard_perl_infix_operator,
							  perl_operator => 'and', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr OROR scalar_expr    { $$ = {  interpret_string => ("OROR(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '||' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_standard_perl_infix_operator,
							  perl_operator => '||', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr ANDAND scalar_expr  { $$ = {  interpret_string => ("ANDAND(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '&&' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_standard_perl_infix_operator,
							  perl_operator => '&&', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '|' scalar_expr     { $$ = {  interpret_string => ("BITOR(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '|' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '|', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '^' scalar_expr     { $$ = {  interpret_string => ("BITXOR(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '^' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '^', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '&' scalar_expr     { $$ = {  interpret_string => ("BITAND(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '&' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '&', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr EQk scalar_expr     { $$ = {  interpret_string => ("EQk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' eq ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'eq', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '=' scalar_expr     { $$ = {  interpret_string => ("EQ(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '=' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'eq', /* TBD special version? */
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr EQEQ scalar_expr    { $$ = {  interpret_string => ("EQEQ(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' == ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '==', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr LTk scalar_expr     { $$ = {  interpret_string => ("LTk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' lt ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'lt', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr LEk scalar_expr     { $$ = {  interpret_string => ("LEk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' le ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'le', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr NEk scalar_expr     { $$ = {  interpret_string => ("NEk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' ne ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'ne', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr GTk scalar_expr     { $$ = {  interpret_string => ("GTk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' gt ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'gt', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr GEk scalar_expr     { $$ = {  interpret_string => ("GEk(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' ge ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'ge', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr LT  scalar_expr     { $$ = {  interpret_string => ("LT(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '<' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '<', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr LE  scalar_expr     { $$ = {  interpret_string => ("LE(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '<=' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '<=', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr GT  scalar_expr     { $$ = {  interpret_string => ("GT(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '>' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '>', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr GE  scalar_expr     { $$ = {  interpret_string => ("GE(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '>=' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '>=', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr NE  scalar_expr     { $$ = {  interpret_string => ("NE(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '!=' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '!=', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr EQ  scalar_expr     { $$ = {  interpret_string => ("EQ(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '=' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => 'eq', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr LSHFT scalar_expr   { $$ = {  interpret_string => ("LSHFT(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '<<' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '<<', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr RSHFT scalar_expr   { $$ = {  interpret_string => ("RSHFT(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '>>' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '>>', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr ARSHFT scalar_expr { $$ = {  interpret_string => ("ARSHFT(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '>>>' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_ERROR,
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '+' scalar_expr     { $$ = {  interpret_string => ("ADD(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '+' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '+', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '-' scalar_expr     { $$ = {  interpret_string => ("SUB(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '-' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '-', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '*' scalar_expr     { $$ = {  interpret_string => ("MUL(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '*' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '*', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '/' scalar_expr     { $$ = {  interpret_string => ("DIV(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '/' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '/', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr '%' scalar_expr     { $$ = {  interpret_string => ("MOD(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '%' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '%', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr DIV scalar_expr     { $$ = {  interpret_string => ("IDIV(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' div ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_ERROR,
							  perl_operator => '', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr REM scalar_expr     { $$ = {  interpret_string => ("REM(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' rem ' . $3->{eval_name}),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
						}
|               scalar_expr MOD scalar_expr     { $$ = {  interpret_string => ("MOD(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' mod ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_ERROR,
							  perl_operator => '', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr MATCH scalar_expr   { $$ = {  interpret_string => ("MATCH(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . ' =~ ' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '=~', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr POW scalar_expr     { $$ = {  interpret_string => ("POW(" . deref($1) . "," . deref($3) . ")"),
							  eval_name => ($1->{eval_name} . '**' . $3->{eval_name}),
							  eval_function => \&eval_functions::eval_null_propagating_perl_infix_operator,
							  perl_operator => '**', 
							  left_operand => $1, right_operand => $3,
							};
						}
|               scalar_expr IN subquery { $$ = {  interpret_string => ("IN(" . deref($1) . "," . deref($3) . ")"),
						  eval_name => ($1->{eval_name} . ' in ' . $3->{eval_name}),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
						}
        # IS NULL and IS NOT NULL handled specially */
        #|      scalar_expr IS table_expr       { $$ = {  interpret_string => ("IS(" . deref($1) . "," . deref($3) . ")"),
        #                                                 eval_function => \&eval_functions::eval_ERROR,
        #                                              };
        #                                       }
        
        # BUG: TBD: NOT IN grammar problems
        #|      scalar_expr NOTk IN table_expr  { $$ = {  interpret_string => ("NOT(IN(" . deref($1) . "," . deref($4) . "))"),
        #                                                 eval_function => \&eval_functions::eval_ERROR,
        #                                              };
        #                                       }
        
;

subquery: '(' table_expr ')' { $$ = {  interpret_string => (deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


complex_table_expr:  select_expr      { $$ = $1; }
|              update_expr      { $$ = $1; }
|              insert_expr      { $$ = $1; }
|              delete_expr      { $$ = $1; }
|              alter_expr      { $$ = $1; }
;


# SELECT is implemented (mainly)

select_expr:   select_expr_formation	{ $$ = $1; }

select_expr_formation:
               # null form
               select_operator { $$ = {  interpret_string => (deref($1) . "()"),
                                         eval_function => \&eval_functions::eval_SELECT,
				         select_operator => $1,  
                                       };
                               }
        
|              select_operator selection   { $$ = {  interpret_string => (deref($1) . "(" . deref($2) . ")"),
						     selection => $2,
						     select_operator => $1,  
						     eval_function => \&eval_functions::eval_SELECT,
                                                  };
					   }
|              select_expr_formation fiw_clause   { $$ = add_fields( $1, { $2->{clause} => $2, } );
					  $$->{interpret_string} = (deref($1) . "/fiw=" . deref($2));
					}
|              select_expr_formation order_by_clause   { $$ = {  interpret_string => (deref($1) . "/" . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

select_operator:    SELECT		{ $$ = {  interpret_string => 'SELECT', operator => 'SELECT', }; }
|		    SELECT quantifier   { $$ = {  interpret_string => 'SELECT_' . deref($2),
						  operator => 'SELECT', quantifier => deref($2),
						};
					}
;

quantifier: ALL { $$ = {  interpret_string => ('ALL'),
                          eval_function => \&eval_functions::eval_ERROR,
                       };
		  die "TBD: quantifiers ALL/DISTINCT not yet implemented\n";
                }
|     DISTINCT  { $$ = {  interpret_string => ('DISTINCT'),
                          eval_function => \&eval_functions::eval_ERROR,
                       };
		  die "TBD: quantifiers ALL/DISTINCT not yet implemented\n";
                }
;


selection: scalar_expr_as_commalist	      { $$ = $1; $$->{interpret_string} = 'LIST(' . $1->{interpret_string} . ')'; }
|	   wildcard			      { $$ = $1; }
        /* TBD: t1.* is broken */
|         identifier_dotlist '.' wildcard     { $$ = {  interpret_string => ('WILDCARD(' . deref($1) . ',' . deref($3) . ')'),
                                                        eval_function => \&eval_functions::eval_ERROR,
                                                     };
					      }
 
;

scalar_expr_as_commalist:  
    scalar_expr_as				{ $$ = { interpret_string => (deref($1) ),
                                                         eval_function => \&eval_functions::eval_ERROR,
							 as_list => [ $1->{as_mapping}, ],
						       };
                                                }
|   scalar_expr_as_commalist ',' scalar_expr_as	{ $$ = {  interpret_string => ( deref($1) . ", " . deref($3) ),
                                                          eval_function => \&eval_functions::eval_ERROR,
							  as_list => [ @{$1->{as_list}}, $3->{as_mapping}, ],
                                                       };
                                                }
;

scalar_expr_as:
    scalar_expr				{ $$ = {  interpret_string => (deref($1) ),
						  as_mapping => { as_name => ($1->{eval_name}||"undef"), 
								    /* TBD: need to synthesize names for expressions */
								  as_value => $1, 
								},
                                                  eval_function => \&eval_functions::eval_ERROR,
						};
					}
|   scalar_expr AS simple_identifier	{ $$ = {  interpret_string => ("AS(" . deref($1) . ", "
								       . deref($3) . ")" ),
						  eval_function => \&eval_functions::eval_ERROR,
						  as_mapping => { as_name => $3->{eval_name},
								  as_value => $1,
								},
                                               };
                                        }
;


table_identifier:   simple_identifier		{ $$ = $1; $$->{type} = { simple_identifier => 1} }
|		    TABLE '(' scalar_expr ')'	{ $$ = $1; } /* TBD: do coercion */
;

column_ref:	    identifier_dotlist  { $$ = $1; $$->{eval_function} = \&eval_functions::eval_column_ref;
					  $$->{is_column_ref} = 1;
					}
;

simple_identifier:	IDENTIFIER		{ $$ = { interpret_string => ("id(" . deref($1) . ")"),
							 eval_function => \&eval_functions::eval_ERROR,
                                                         eval_name => deref($1),
						       };
						}
|                       DELIMITED_IDENTIFIER    { $$ = {  interpret_string => ("d_id(" . deref($1) . ")"),
							  eval_function => \&eval_functions::eval_ERROR,
                                                          eval_name => eval deref($1),
                                                       };
						}
;
identifier_dotlist:
    simple_identifier				{ $$ = {  column_ref_components => [ $1->{eval_name} ],
							  eval_name => $1->{eval_name},
							  eval_function => \&eval_functions::eval_column_ref,
							  interpret_string => deref($1),
						       };
						}
|   identifier_dotlist '.' simple_identifier	{ $$ = {  interpret_string => (deref($1) . '.' . deref($3)),
							  eval_name => $1->{eval_name} . '.' . $3->{eval_name},
							  column_ref_components => [ @{$1->{column_ref_components}},
										     $3->{eval_name} ],
						       };
                                                }
;

# TBD: UPDATE is not implemented yet, although present in grammar
update_expr :   ag_update_expression { $$ = $1; }
|     standard_update_expression { $$ = $1; }
;

ag_update_expression:
    UPDATE { $$ = {  interpret_string => ('UPDATE'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|   ag_update_expression SET assignment_commalist { $$ = {  interpret_string => (deref($1) . '/set=(' . deref($3) . '))'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|   ag_update_expression fiw_clause { $$ = {  interpret_string => (deref($1) . '/' . deref($2) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


standard_update_expression:
    UPDATE table_identifier SET assignment_commalist { $$ = {  interpret_string => ('UPDATE(table=' . deref($2) . ',set=(' . deref($4) . '))'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|   UPDATE table_identifier SET assignment_commalist where_clause
        { $$ = {  interpret_string => ('UPDATE(table=' . deref($2) . ',set=(' . deref($4) . '))/WHERE=' . deref($5)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

assignment_commalist: assignment { $$ = $1; }
|       assignment_commalist ',' assignment { $$ = {  interpret_string => (deref($1) . ',' . deref($3) ),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

simple_column_name: simple_identifier;

assignment: simple_column_name EQ scalar_expr { $$ = {  interpret_string => ('SET(' . deref($1) . ',' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


extended_assignment_commalist: extended_assignment { $$ = $1; }
|       extended_assignment_commalist ',' extended_assignment { $$ = {  interpret_string => (deref($1) . ',' . deref($3) ),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

extended_assignment: simple_column_name extended_assignment_op scalar_expr { $$ = {  interpret_string => ('SET(' . deref($1) . ',' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;
extended_assignment_op:
    EQ { $$ = {  interpret_string => ("SET('=')"),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|     SET_COLON_EQUALS { $$ = {  interpret_string => ("SET(':=')"),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|     SET_RIGHT_DOUBLE_ARROW { $$ = {  interpret_string => ("SET('=>')"),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


# TBD: ALTER TABLE ... DROP COLUMNS ... is not implemented yet, although present in grammar
alter_expr:     ALTER TABLE			{ $$ = { interpret_string => ("ALTER_TABLE"),
							 eval_function => \&eval_functions::eval_ERROR,
                                                       };
                                                }
|               ALTER TABLE table_identifier	{ $$ = { interpret_string => ("ALTER_TABLE/table=" . deref($3)),
							 eval_function => \&eval_functions::eval_ERROR,
							 FROM => $3,
							 INTO => $3,  
                                                       };
					        }
|               alter_expr fiw_clause           { $$ = add_fields( $1, { $2->{clause} => $2, } );
					          $$->{interpret_string} = (deref($1) . "/fiw=" . deref($2));
					        }
|               alter_expr DROP COLUMN column_commalist
						{
						    $$ = $4;
						    $$->{interpret_string} = (deref($1) . "/drop_column=(" . deref($4) . ")" );
                                                    #$$->{eval_function} = \&eval_functions::eval_ALTER_TABLE_DROP_COLUMN;
                                                }
;


# TBD: INSERT is not implemented yet, although present in grammar
insert_expr:    INSERT                  { $$ = {  interpret_string => ('INSERT'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|               insert_expr fi_clause   { $$ = {  interpret_string => (deref($1) . '/' . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|               insert_expr select_expr { $$ = {  interpret_string => (deref($1) . '/select=' . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|               insert_expr values { $$ = {  interpret_string => (deref($1) . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
    #debug |    insert_expr table_expr { $$ = {  interpret_string => (deref($1) . deref($2)),
    #                                            eval_function => \&eval_functions::eval_ERROR,
    #                                         };
    #                                  }
    #bug: when I enable INSERT table_expr, INSERT simple_column_name VALUES values  fails
;

values: VALUES '(' scalar_expr_commalist ')'  { $$ = {  interpret_string => ('/values=(' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|       VALUES '{' extended_assignment_commalist '}'  { $$ = {  interpret_string => ('/values_set=(' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|       VALUES '(' extended_assignment_commalist ')'  { $$ = {  interpret_string => ('/values_set=(' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
|       '(' column_commalist ')' VALUES '(' scalar_expr_commalist ')'
                                    { $$ = {  interpret_string => ('/value_names=(' . deref($2) . ')/values=(' . deref($6) . ')'),
                                              eval_function => \&eval_functions::eval_ERROR,
                                            };
                                    }
    # TBD: add a Perl syntax?
;



scalar_expr_commalist:  
    scalar_expr					{ $$ = { interpret_string => (deref($1) ),
							 eval_name => $1->{eval_name},
                                                         eval_function => \&eval_functions::eval_LIST,
							 list => [ $1, ],
						       };
                                                }
|   scalar_expr_commalist ',' scalar_expr	{ $$ = {  interpret_string => ( deref($1) . ", " . deref($3) ),
	    						  eval_name => $1->{eval_name} . ',' . $3->{eval_name},
                                                          eval_function => \&eval_functions::eval_LIST,
							  list => [ @{$1->{list}}, $3, ],
                                                       };
                                                }
;



column_commalist: simple_column_name { $$ = $1; }
|       column_commalist ',' simple_column_name { $$ = {  interpret_string => (deref($1) . ',' . deref($3) ),
                                                          eval_function => \&eval_functions::eval_ERROR,
                                                       };
                                                }
;



# DELETE is implemented (mainly)
delete_expr:    DELETE			{ $$ = {  interpret_string => ("DELETE"),
                                                  eval_function => \&eval_functions::eval_DELETE,
                                               };
					}
|               delete_expr fiw_clause  { $$ = add_fields( $1, { $2->{clause} => $2, } );
					  $$->{interpret_string} = (deref($1) . "/fiw=" . deref($2));
					}
;



fi_clause:     from_clause { $$ = $1; }
|              into_clause { $$ = $1; }
;
fiw_clause:    from_clause { $$ = $1; }
|              into_clause { $$ = $1; }
|              where_clause { $$ = $1; }
;

from_clause:    FROM table_expr_as    { $$ = { interpret_string => ('FROM(' . deref($2) . ')'),
					       eval_function => \&eval_functions::eval_ERROR,
					       clause => 'FROM',
					       source => $2, 
                                             };
                                      }
# disabled: subsumed by the above
#|      FROM join_table_expression    { $$ = {  interpret_string => ('FROM(' . deref($2) . ')'),
#                                                eval_function => \&eval_functions::eval_ERROR,
#                                            };
#                                      }
;


order_by_clause:  ORDER BY ordering_spec_commalist     { $$ = {  interpret_string => ('ORDER_BY(' . deref($3) . ')'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;


ordering_spec_commalist: ordering_spec { $$ = $1; }
|                        ordering_spec_commalist ',' ordering_spec  { $$ = {  interpret_string => (deref($1) . "," . deref($3)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;

ordering_spec:  column_ref  { $$ = $1; }
|               column_ref  order_direction_clause { $$ = {  interpret_string => (deref($1) . '/direction=' . deref($2)),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
}
;
order_direction_clause: 
      ASCENDING  { $$ = {}; $$->{interpret_string}="ASCENDING"; }
|     DESCENDING { $$ = {}; $$->{interpret_string}="DESCENDING"; }
;


where_clause:  WHERE scalar_expr     { $$ = {  interpret_string => ('WHERE(' . deref($2) . ')'),
                                               eval_function => \&eval_functions::eval_ERROR,
					       clause => 'WHERE', 
					       predicate => $2,  
                                            };
                                     }
;

into_clause:   INTO table_identifier   { $$ = {  interpret_string => ('INTO(' . deref($2) . ')'),
                                                 eval_function => \&eval_functions::eval_ERROR,
						 clause => 'INTO',
						 target => $2,  
                                               };
                                       }
;

join_table_expression:
    table_expr_as join_op table_expr_as join_on_clause
	{ $$ = {};
	  $$->{interpret_string} = deref($2) . "(" . deref($1) . "," . deref($3) . ")" . "/ON(" . deref($4) . ")";
	  $$->{eval_function} = \&eval_functions::eval_ERROR;
	  $$->{left} = $1;
	  $$->{right} = $3;
	  $$->{join_op} = deref($2);
	  $$->{on} = $4;
        }
|  table_expr_as join_op table_expr_as { die "'table JOIN table' without an ON clause not yet supported\n"; }
;

table_expr_as:
    table_expr			    { $$ = { table_expr => $1,
				             interpret_string => deref($1),
					     #correlation_name => $1->{eval_name},
                                             eval_function => \&eval_functions::eval_ERROR,
					   }
				    }
|   table_expr AS simple_identifier { $$ = { table_expr => $1,
					     correlation_name => $3->{eval_name},
					     interpret_string => ("AS(" . deref($1) . ", " . deref($3) . ")"),
                                             eval_function => \&eval_functions::eval_ERROR,
                                           };
                                    }
;


join_op:          JOIN  { $$ = {  interpret_string => ('JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 INNER JOIN { $$ = {  interpret_string => ('INNER_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 OUTER JOIN { $$ = {  interpret_string => ('OUTER_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 LEFT JOIN { $$ = {  interpret_string => ('LEFT_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 RIGHT JOIN { $$ = {  interpret_string => ('RIGHT_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 LEFT OUTER JOIN { $$ = {  interpret_string => ('LEFT_OUTER_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 RIGHT OUTER JOIN { $$ = {  interpret_string => ('RIGHT_OUTER_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
|                 FULL OUTER JOIN { $$ = {  interpret_string => ('FULL_OUTER_JOIN'),
                                                  eval_function => \&eval_functions::eval_ERROR,
                                               };
                        }
;

# Note that "t1 JOIN t2 ON single_field" is legitimate (shared field)
join_on_clause:   ON scalar_expr_as		    { $$ = $2; }
|		  USING '(' simple_column_name ')'
		    { die "perl-sql does not yet support JOIN/USING. the nonstard JOIN/ON/AS clause can be used instead\n";
		    }
|		  USING '(' column_commalist ',' simple_column_name ')'
		    { die "perl-sql does not yet support multicolumn JOIN USING ( .. ) clauses\n"; }
;

%%
# $Id: ag_perl_sql_parser.y,v 1.1.1.1 1998/05/05 23:52:23 glew Exp $


sub yylex
{
    my ($lex_stream) = @_;
    my ($c, $val);


    my $token = $lex_stream->get_token;

    if( !defined($token) ) {
        return 0;
    }
    elsif( !defined $token->{token} ) {
        die 'Unknown token type: $token->{token} undefined\n'
            . "in yylex (converting from dumb_line_re_lexer tokens to byacc tokens for SQL)\n"
            . Data::Dumper->Dump([$token],['token'])
;
    }
    elsif( $token->{token} eq 'self' ) {
        return ord($token->{string});
    }
    else {
        return ($token->{token}, $token);
    }

}


sub yyerror {
    my ($msg, $s) = @_;
    die "$msg" . ".\n";
}


sub deref {
    my($arg) = @_;

    if( defined $arg->{string} ) {
        return $arg->{string};
    }
    elsif( defined $arg->{interpret_string} ) {
        return $arg->{interpret_string};
    }
    else {
        Carp::confess "unknown structure in parser: "
          . Data::Dumper->Dump([$arg],['deref'])
          . "\n";
    }
}


sub add_fields {
    my($to_hash_ref, $add_hash_fields_ref) = @_;

    my $k;
    foreach $k ( keys %$add_hash_fields_ref ) {
	if( exists $to_hash_ref->{$k} ) {
	    Carp::confess "attempting to redefine already defined field $k\n"
		. Data::Dumper->Dump([$to_hash_ref],['adding-to'])
		. Data::Dumper->Dump([$to_hash_ref],['adding-fields'])
		;
	}
	$to_hash_ref->{$k} = $add_hash_fields_ref->{$k};
    }
    return $to_hash_ref;
}


1;
