#!/s/perl-5.004/bin/perl -w

## $Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/db_algorithms.pm,v 1.2 1998/05/06 02:20:53 glew Exp $

##
## db_algorithms.pm
##
## database algorithms for the plain_text_database
## (although they are being written so that they might be useable elsewhere)
## 
## The present algorithms are very basic, and will consist only of
## (a) whole file scans
## (b) algorithms that take advantage of sorted files
##     (b.1) sort-merge joins
##     (b.2) possibly some seek based scans
##           using estimates of location in sorted file
## Sure, it'll be slow, but it's better than nothing
## (and one day may be extended to handle indexes
## - but always in plain text files).
## 

## TBD: somehow I want to make this object like,
## but it seems unnatural. 
## plain_text_database may need to inherit from here.
## but not in general, since a query optimizer will
## have to pick and choose - i.e. inheritance doesn't
## seem quite right.

use strict;

package db_algorithms;

# Deduce location of libraries distributed with package
use FindBin;
BEGIN
{
    my $install_root = $FindBin::RealBin;
    $install_root =~ s:(.*/perl-sql)/.*:$1:;
    unshift (@INC, "$install_root/lib");
}

use Perl_SQL::plain_text_database;

use Carp;
use English;
use Data::Dumper;
use IO::File;
use IO::Seekable;
use strict;
#use diagnostics;


my $parse_errors_death = 1;	# set to 0 if want to ignore parse errors (e.g. in test)
my $ParseErrorString = "special_error_string_returned"
    . "_if_a_parse_error_is_detected"
    . "_but_you_do_not_want_to_die_deep_down_in_recursion"
    . "_and_want_to_return_a_better_full_error_message";

$db_algorithms::evaluate_table_and_field_specifier = sub {
    die "variable evaluated without a database context\n";
};


# eval_string_or_lambda
#
# evaluate an expression passed in either as a string to be eval'ed,
# or as a lambda function.
#
# Background: the original prototype implemented "SELECT ... WHERE conditions"
# and "SELECT selection_lists" using strings that were transformed to use
# internal functions for variable dereferencing, etc., and then eval'ed.
#
# This was cumbersome for handling complex expressions, so Perl-SQL instead
# specifies these things using lambda functions.
#
# The eval's may, in fact, be faster than interpreting data structures
# using lambda functions...
#
# TBD: not used because expressions required access to lexically scoped variables.
# TBD: fix.
#
#sub eval_string_or_lambda {
#    my $arg = @_;
#
#    if( ref $arg eq 'CODE' ) {
#	&{$arg}();
#    }
#    else {
#	eval $arg;
#    }
#}
#

sub scan_select {
    my %args = @_;

    (my $db = $args{'database'}) || confess "scan_select requires an input *database*";
    (my $outdb = $args{'output'}) || confess "scan_select requires an *output* database";

    my $original_condition_string = (exists $args{'where'} ? $args{'where'} : 1);
    my $select_condition = parse_expression($original_condition_string);
    if( $select_condition =~ m/$ParseErrorString/ ) {
	die "error parsing select condition\n";
    }

    my %selection = ();
    my %parsed_selection = ();
    my $select_all = 0;
    if( $args{'select'} eq '*' ) {
	%selection = ();
	$select_all = 1;
    }
    elsif( $args{'select'} ) {
	%selection = %{$args{'select'}};
	$select_all = 0;	# may repair later
	if( ! %selection ) {
	    %selection = ();
	    $select_all = 1;
	}
    }
    else {
	# no selection - take all;
	%selection = ();
	$select_all = 1;
    }
    foreach my $label ( keys %selection ) {
	$parsed_selection{$label} = parse_expression($selection{$label});;
	if( $parsed_selection{$label} =~ m/$ParseErrorString/ ) {
	    die "error parsing select condition\n";
	}
    }
    

    my($InputTupleRef);
    local $db_algorithms::evaluate_table_and_field_specifier # NB: dynamic scoping via local, so that outside lambda function can call it
	= sub { 
	    my $arg = shift;
	    if( @_ ) { die "too many arguments / bad expression / for a single table join (no correlation names)\n"; }
	    my $val = $$InputTupleRef{$arg}; # TBD: do validity checking
	    if( !defined $val ) { return undef; }
	    return $val;
	};

    if( $args{selection_all_aggregates} ) {
	$InputTupleRef = {};
	if( $select_all ) {
	    $outdb->put_next_tuple( $InputTupleRef );
	}
	else {
	    my %OutputTuple = ();
	    foreach my $label ( sort keys %selection ) {
		#disable diagnostics;
		my $value = (ref  $parsed_selection{$label}  eq 'CODE' 
			     ? &{$parsed_selection{$label}}() 
			     : eval  $parsed_selection{$label}  );
		#enable diagnostics;
		$OutputTuple{$label} = $value;
	    }
	    $outdb->put_next_tuple( \%OutputTuple );
	}
    }
    else {
	while( $InputTupleRef = $db->get_next_tuple() ) {
	    #disable diagnostics;
	    my $sc = (ref  $select_condition  eq 'CODE' ? &{$select_condition}() : eval  $select_condition  );
	    #enable diagnostics;
	    if( $sc ) {
		if( $select_all ) {
		    $outdb->put_next_tuple( $InputTupleRef );
		}
		else {
		    my %OutputTuple = ();
		    foreach my $label ( sort keys %selection ) {
			#disable diagnostics;
			my $value = (ref  $parsed_selection{$label}  eq 'CODE' ? &{$parsed_selection{$label}}() : eval  $parsed_selection{$label}  );
			#enable diagnostics;
			$OutputTuple{$label} = $value;
		    }
		    $outdb->put_next_tuple( \%OutputTuple );
		}
	    }
	}
    }

}






# perform comparisons for a join,
# according to number or string
# returns -1,0,+1 for the usual directions,
# but also left_null(-2), right_null(+2), both_null(-99)
# to facilitate null handling.
# Note that the sign indicates where to advance
# - advancing left if <0 is always okay.
# TBD: handle exists?

# TBD: factor ouut the number/string part for efficiency
# - I coded it this way for ease.

my $join_cmp_both_null = -99;
my $join_cmp_left_null = -2;
my $join_cmp_right_null = +2;

sub join_cmp {
    my %args = @_;		# need this, since may have undef values to compare

    my $left = $args{'left'};
    my $right = $args{'right'};
    my $number = $args{'number'};

    if( !defined $left && !defined $right ) {
	return $join_cmp_both_null;
    }
    elsif( !defined $left ) {
	return $join_cmp_left_null;
    }
    elsif( !defined $right ) {
	return $join_cmp_right_null;
    }
    elsif( $number ) {
	return $left <=> $right;
    }
    else {
	return $left cmp $right;
    }
}



sub sorted_merge_join {
    my %args = @_;

    ## Argument handling
    
    # input databases
    if( !defined $args{'left'}  or !defined $args{'right'} ) {
	confess "sorted_merge_join requires 2 input databases, *left* and *right*.";
    }
    my $db_left = $args{'left'};
    my $db_right = $args{'right'};

    my $correlation_names = $args{'correlation_names'};

    # output database
    (my $outdb = $args{'output'}) || confess "sorted_merge_join requires an *output* database";

    # on condition

    # TBD: eventually will specify an ON condition for joins
    # but for now will specify the pieces I need individually
    #TBD: my $original_on_condition_string = $args{'join_on_condition'} || 1;
    #TBD: my $parsed_on_condition = parse_expression($original_on_condition_string);
    #TBD: my $merge_on_variable = determine_merge_variable($parsed_on_condition_string);
    # TBD: join_on_label from %args, or from fields if names same, or ...

    my $join_cmp = $args{'join_cmp'} || "eq";
    my $join_cmp_number;
    if( $join_cmp eq "==" ) {
	$join_cmp_number = 1;
    }
    elsif( $join_cmp eq "eq" ) {
	$join_cmp_number = 0;
    }
    else {
	die "unknown join_cmp = $join_cmp. require == or eq\n";
    }


    my $join_on_left_var = $args{'join_on_left_var'} || die "must define join_on_left_var\n";
    my $join_on_right_var = $args{'join_on_left_var'} || die "must define join_on_right_var\n";
    my $join_on_var_label;
    if( defined $args{'join_on_var_label'} ) {
	$join_on_var_label = $args{'join_on_var_label'};
    }
    elsif( $join_on_left_var eq $join_on_right_var ) {
	$join_on_var_label = $join_on_left_var;
    }
    else {
	die "must define join_on_var_label, or join on identically named fields\n";
    }



    


    # type of join
    my $outer_left = 0;
    my $outer_right = 0;
    if( $args{'join'} eq 'inner' || $args{'join'} eq 'INNER_JOIN' ) {
	$outer_left = 0;
	$outer_right = 0;
    }
    elsif( $args{'join'} eq 'outer-left' || $args{'join'} eq 'LEFT_OUTER_JOIN' ) {
	$outer_left = 1;
	$outer_right = 0;
    }
    elsif( $args{'join'} eq 'outer-right' || $args{'join'} eq 'RIGHT_OUTER_JOIN' ) {
	$outer_left = 0;
	$outer_right = 1;
    }
    elsif( $args{'join'} eq 'outer-full' || $args{'join'} eq 'FULL_OUTER_JOIN' ) {
	$outer_left = 1;
	$outer_right = 1;
    }
    else {
	die "unknown join type $args{'join'}\n"
	    . "supported types are inner/_left outer/_right outer/full outer\n";
	# TBD: support NATURAL JOIN, CROSS JOIN (which are not algorithms,
	# but more merge criteria)
    }
   

    # where condition
    my $original_where_condition_string = (exists $args{'where'} ? $args{'where'} : 1);
    my $parsed_where_condition = parse_expression($original_where_condition_string);
    if( $parsed_where_condition =~ m/$ParseErrorString/ ) {
	die "error parsing select condition\n";
    }



    # select expressions and labels
    my %selection = ();
    my %parsed_selection = ();
    my $select_all = 0;
    if( !defined $args{'select'} || $args{'select'} eq '*' ) {
	%selection = ();
	$select_all = 1;
    }
    elsif( $args{'select'} ) {
	%selection = %{$args{'select'}};
	$select_all = 0;	# may repair later
	if( ! %selection ) {
	    %selection = ();
	    $select_all = 1;
	}
    }
    else {
	# no selection - take all;
	%selection = ();
	$select_all = 1;
    }
    foreach my $label ( keys %selection ) {
	$parsed_selection{$label} = parse_expression($selection{$label});;
	if( $parsed_selection{$label} =~ m/$ParseErrorString/ ) {
	    die "error parsing select condition\n";
	}
    }

    # end of argument parsing

    #########################################################################

    # join algorithm setup
    
    
    my($t_left,$t_right);

    my $tuple_left = $db_left->get_next_tuple();
    my $tuple_right = $db_right->get_next_tuple();

    my $last_left_tuple = ();
    my $last_right_tuple = ();

    # some helper functions to make the algorithm more readable and less repetitious
    my $advance_join_left = sub {
	$last_left_tuple = $tuple_left;
	$tuple_left = $db_left->get_next_tuple();
    };
    my $advance_join_right = sub {
	$last_right_tuple = $tuple_right;
	$tuple_right = $db_right->get_next_tuple();
    };

    my $check_sorted = sub {
	# check that file is sorted
	# TBD: extend to do equijoins on multiple keys, by looping
	if( 0 <= join_cmp( left => $last_left_tuple->{$join_on_left_var},
			  right => $tuple_right->{$join_on_left_var}, 
			  number => $join_cmp_number )	   
	   ) {
	       die "error: left join table has repeated or unordered keys\n"
		   . "near key = " . ($last_left_tuple||"null") . ", " . ($tuple_right->{$join_on_right_var}||"null") . "\n"
		   . "(restriction: can only handle sorted files with unique keys)\n";
	   }
	if( 0 <= join_cmp( left => $last_right_tuple->{join_on_right_var},
		     right => $tuple_right->{$join_on_right_var}, 
		     number => $join_cmp_number )
	   ) {
	       die "error: right join table has repeated or unordered keys\n"
		   . "near key = " . ($last_right_tuple||"null") . ", " . ($tuple_right->{$join_on_right_var}||"null") . "\n"
		   . "(restriction: can only handle sorted files with unique keys)\n";
	   }
       };


    # join algorithm main loop

    for(;;) {

	my $left_exists = (defined $tuple_left && %$tuple_left);
	my $right_exists = (defined $tuple_right && %$tuple_right);

	if( !$left_exists && !$right_exists ) { last; }

	# basic join stuff

	&$check_sorted();

	
	my $join_cmp_result;
	if( !$left_exists ) {
	    $join_cmp_result = +99; # force advance right
	}
	elsif( !$right_exists ) {
	    $join_cmp_result = -99; # force advance left
	}
	else {
	    $join_cmp_result = join_cmp( left => $tuple_left->{$join_on_left_var}, 
				       right => $tuple_right->{$join_on_right_var}, 
				       number => $join_cmp_number );
	}


	# The actual join comparison:
	# (0) the classic merging IF statement: =, <, >
	# (1) construct the potentially_merged_tuple, which contains fields
	#     known to be common to both sides, for which no table qualifier is necessary
	#     (for simple equijoins, this is only the join'ed on variable)
        # (2) execute WHERE and SELECT to fully merge the tuples

	if( $join_cmp_result == 0 ) {
	    # Note: this case only handles exact eq or == equality, no nulls
	    # TBD: it might be useful to provide an XOR JOIN - specifically excluding the common matched elements
	    my $result = merge_where_select( merged => { $join_on_var_label => $tuple_left->{$join_on_left_var}, },
					    left=>$tuple_left, right=>$tuple_right,
					    where=>$parsed_where_condition,
					    'select'=>($select_all ? '*' : \%parsed_selection),
					    correlation_names=>$correlation_names,
					    );
	    if( $result ) {
		$outdb->put_next_tuple( $result );
	    }

	    &$advance_join_left();
	    &$advance_join_right();
	    # TBD: currently erroring if multiple tuples with same join key; should loop
	}
	elsif(  $join_cmp_result < 0 ) {
	    # Note: this case handles lt, <, but also left_null and both_null for the ON variables
	    if( $outer_left ) {
		my $result = merge_where_select( merged => { (!exists $tuple_left->{$join_on_left_var} ? () 
							   : ($join_on_var_label =>  $tuple_left->{$join_on_left_var}), ),
						      },
					     left=>$tuple_left, right=>{},
					     where=>$parsed_where_condition,
					     'select'=>($select_all ? '*' : \%parsed_selection),
					     correlation_names=>$correlation_names,
					     );
		if( $result ) {
		    $outdb->put_next_tuple( $result );
		}
	    }
	    &$advance_join_left();
	}
	elsif( $join_cmp_result > 0 ) {
	    # Note: this case handles lt, <, but also right_null for the ON variables
	    if( $outer_right ) {
		my $result = merge_where_select( merged => { (!exists $tuple_right->{$join_on_right_var} ? () 
							   : ($join_on_var_label =>  $tuple_right->{$join_on_right_var}) ),
						      },
					     left=>{}, right=>$tuple_right,
					     where=>$parsed_where_condition,
					     'select'=>($select_all ? '*' : \%parsed_selection),
					     correlation_names=>$correlation_names,
					     );
		if( $result ) {
		    $outdb->put_next_tuple( $result );
		}
	    }
	    &$advance_join_right();
	}
	else {
	    die "unexpected JOIN condition\n";
	}
    }
}



sub merge_where_select {
    # evaluate the WHERE tuple reduction clause
    # and the SELECT field extraction clauses
    # in an environment where 
    # the original tuples can be accessed by correlation name (e.g. LEFT, RIGHT)
    # and some fields can be accessed without a correlation name
    # (those passed in in the merged tuple, basically the equijoin field)

    my %arg = @_;

    my $merged_tuple = $arg{'merged'} || {};
    my $left_tuple = $arg{'left'} || {};
    my $right_tuple = $arg{'right'} || {};

    my $where_condition = (exists $arg{'where'} ? $arg{'where'} : 1);
    my $select = $arg{'select'} || 1;

    my $correlation_names = $arg{'correlation_names'};

    # For now I will use LEFT and RIGHT as my correlation names
    # TBD: pass proper names (tablenames and correlation names) in.



    local $db_algorithms::evaluate_table_and_field_specifier # NB: dynamic scoping via local, so that outside lambda function can call it
	= sub { 
	    my ($first,$second,$extra) = @_;
	    if( $extra ) { die "too many arguments / bad expression. Only supports field, table.field\n"; }
	    
	    if( !defined $second ) {
		# TBD: use real correlation names, in addition to merged names
		if( !exists $merged_tuple->{$first} ) {
		    return undef;
		}
		else {
		    return $merged_tuple->{$first};
		}
	    } 
	    else {
		my $cn = $correlation_names->{$first};
		if( !defined $cn ) {
		    die "bad correlation name value $first.$second\n";
		}
		elsif( $cn == -1 ) {
		    return $left_tuple->{$second};
		}
		elsif( $cn == +1 ) {
		    return $right_tuple->{$second};
		}
		else {
		    die "unexpected correlation name value $first.$second = $cn\n";
		}
	    }
	};

    #disable diagnostics;
    my $wc =  (ref  $where_condition  eq 'CODE' ? &{$where_condition}() : eval  $where_condition  );
    #enable diagnostics;
    if( $wc ) {
	# TBD: by evaluating the where condition before the merger of the tuples,
	# we lose the ability to use the fieldnames of the actually merged record
	# (the SELECT expre AS fieldnames) in the WHERE clause.  But, we can use
	# the WHERE clause to protect against things like zero divide.
	# TBD: what is the official behaviour?
	if( $select eq '*' ) {
	    # merge both tuples
	    # TBD: uniqify field names?
	    # TBD: currently uniqifying by adding a fixed prefix "LEFT-" or "RIGHT"
	    # TBD: use the proper correlation name
	    # TBD: optionally uniqify only non-unique/non-identical fields
	    my %OutputTuple = ();
	    foreach my $label ( keys %$left_tuple ) {
		$OutputTuple{ ($correlation_names->{PRIMARY_left_table_correlation_name}||"LEFT") ."." .$label} = $left_tuple->{$label};
	    }
	    foreach my $label ( keys %$right_tuple ) {
		$OutputTuple{ ($correlation_names->{PRIMARY_right_table_correlation_name}||"RIGHT") .".".$label} =  $right_tuple->{$label};
	    }
	    return \%OutputTuple;
	}
	else {
	    my %OutputTuple = ();
	    foreach my $label ( sort keys %$select ) {
		#disable diagnostics;
		my $value = (ref  $select->{$label}  eq 'CODE' ? &{$select->{$label}}() : eval  $select->{$label}  );
		#enable diagnostics;
		$OutputTuple{$label} = $value;
	    }
	    return \%OutputTuple;
	}
    }
    else {
	return undef;
    }
}



#FUTURE?: sub determine_merge_variable {
#FUTURE?: 
#FUTURE?:     # determine merge variable by evaluating the merge_on_condition
#FUTURE?:     # in an isolated environment, where the methods that evaluate the variables
#FUTURE?:     # record their names.
#FUTURE?: 
#FUTURE?:     # PRELIMINARY: error if more than one variable per table.
#FUTURE?:     # (could handle by sorting on a composite key)
#FUTURE?:     
#FUTURE?:     die "TBD: Not yet implemented\n";
#FUTURE?: 
#FUTURE?:     
#FUTURE?: }


#debug:sub EvaluateExpressionVariable {
#debug:    my($string) = @_;
#debug:
#debug:    return $$InputTupleRef{$string};
#debug:}


# TBD: this parser is getting out of hand!
# It is becoming a kluged recursive descent parser,
# when originally it was supposed to perform minimal massaging and pass to Perl eval.
# Replace by yacc/lex


my $quote1_re = '\'([^\'\\\\]*(\\.[^\'\\\\]*)*)\'';
my $quote2_re = '\"([^\"\\\\]*(\\.[^\"\\\\]*)*)\"';



sub parse_expression
{
    my ($query) = @_;

    if( !defined $query ) { 
	$query = ""; 
    }

    if( ref $query eq 'CODE' ) {
	return $query; # TBD: don't parse lambda functions passed from Perl-SQL Yacc
    }
    if( ref $query ne '' || ref \$query ne 'SCALAR' ) {	
      Carp::confess "Internal error: parse_expression called with non-scalar (expected string)\n"
	  . Data::Dumper->Dump([\$query],['parse-expression(query)']);
    }


    my $result = breakup_quotes_and_parse( string => $query );

    if( $result =~ m/$ParseErrorString/ ) {
	# TBD: arrange to have recursive children tell us what problem is?
	warn "error parsing expression $query";
	return undef;
    }

    # after non-quotes have been processed,
    # all table/field expressions are of the form
    # 'T1','F1' ...
    # i.e. fully single quoted, with commas.
    # now fix up to be a call

    # TBD: I forget exactly what SQL requires for a fieldname.
    # I know that quotes are allowed, e.g. T1."A+B"
    # (making it easy to create labels without AS clauses),
    # but I forget if double or single.
    # I will assume single for fieldnames,
    # and assume doubles for strings.


    my $fieldname_re = "($quote1_re)";

    my $fieldselector_re = "($fieldname_re(\s*,\s*$fieldname_re)*)";


    $result =~ s/$fieldselector_re/ ' &$db_algorithms::evaluate_table_and_field_specifier( ' .  $1 . '  ) ' /ge;

    return $result;
}




sub breakup_quotes_and_parse {
    my %arg = @_;		# hash, so that can provide a level keyword

    if( !defined $arg{'string'} ) {
	return "";
    }
    elsif( $arg{'string'} =~ m/($quote1_re|$quote2_re)/ ) {
	my $before = $`;
	my $quote = $1;
	my $after = $';
	my $result = "";
	if( defined $before ) {
	    $result .= breakup_quotes_and_parse( string => $before );
	}
	$result .= $quote;
	if( defined $after ) {
	    $result .= breakup_quotes_and_parse( string => $after );
	}
	return $result;
    }
    else {
	return parse_nonquote( $arg{'string'} );
    }
}

sub parse_nonquote {
    my($query) = @_;

    my $raw_query = $query;

    my $dangerous_perl_syntax_regexp = '[`;@$%&]';
    my $filtered_query = $query;
    $filtered_query =~ s/&&/ /g; # allow use of && and
    if( $filtered_query =~ m/$dangerous_perl_syntax_regexp/ ) {
        warn "error: ag's perl sql only accepts limited expressions,\n"
	    . "rejecting expressions matching or containing $dangerous_perl_syntax_regexp.\n"
	    . "rejecting: <$query>\n";
	if( $parse_errors_death ) {
	    die;
	}
	return $ParseErrorString;
    }

    my $identifier_re = "([A-Za-z_][A-Za-z_0-9]*)";

    $query =~ s/$identifier_re/ 
	# don't mess with operator name reserved words
	# TBD: add a more complete set
	my %is_reserved_word = (
			     "and" => 1,
			     "or" => 1,
			     "null" => 1,
			     "not" => 1,
			     "defined" => 1,
			     "exists" => 1,
			     # kluge: pass through Perl operator names,
			     # so that we don't need to worry about field types
			     # (instead, user does)
			     "ge" => 1,
			     "le" => 1,
			     "eq" => 1,
			     "ne" => 1,
			     "lt" => 1,
			     "gt" => 1,
			     );
	if( $is_reserved_word{$1} ) { "$1"; }
        # pass names on
        else { "\'$1\'"; } 
    /gex;
    $query =~ s/\./,/g;
    

    # Regexps - allow Perl regexp syntax
    # TBD: use SQL's more limited regexps,
    # or use MySQL and MiniSQL's extended version?
    # TBD: use a match bracket other than /
    $query =~ s("(m{0,1})/([^/\\]*(\\.[^/\\]*)*)/(i{0,1})")(
        if  ($2)    { "$1/$2/$4"; }
    )gex;


    # Convert SQL operators to Perl operators
    # TBD: really need field types if to do this properly
    # Heuristic: for now, justy let user specify Perl operator
    # TBD: do better than this?
    #$query =~ s/>=\s*(")/ ge $1/g;
    #$query =~ s/<=\s*(")/ le $1/g;
    #$query =~ s/=\s*(")/ eq $1/g;
    #$query =~ s/<>\s*(")/ ne $1/g;
    #$query =~ s/<\s*(")/ lt $1/g;
    #$query =~ s/>\s*(")/ gt $1/g;

    #$query =~ s/([^><!])=([^~])\s*/$1 == $2/g;
    #$query =~ s/<>/ != /g;

    return ($query);
}





###############################################################
# Handle aggregate functions by taking a zero-th pass over the data
# TBD: kluge: really should handle expressions and temporaries cleanly
#      by doing all setup for selection, creating another set of trees
#      describing temporary file creation, etc., optimizing that, etc.
#      But I'm lazy, and can hack it in here.
# TBD: special handling required for non-rewindable files
# TBD: could save work by saving WHERE filtered result in aggregate pass.


sub calculate_aggregates {
    my %arg = @_;

    my $dbfile = $arg{dbfile} || Carp::confess "internal error: need dbfile\n";
    my $filehandle = $dbfile->{NICE_IO_FILEHANDLE}->{NICE_IO_FILEHANDLE};
    my $aggregate_list = $arg{aggregates} || Carp::confess "internal error: need aggregates\n";

    my $where = defined($arg{where}) ? $arg{where} : die "internal error: calculate_aggregates needs where argument\n";
    my $parsed_where = parse_expression($where);
    if( $parsed_where =~ m/$ParseErrorString/ ) {
	die "error parsing select condition\n";
    }
    


    # TBD: currently handling this scan by scanning and then rewinding.
    #    Really should allow copy
    #    Try to detect if non-rewindable

    sub seekdie {
	die "error: need to be able to seek(rewind) to calculate aggregates\n";
    }


    #print "debug: trying to seek for aggregate calculation\n";
    $dbfile->seek(0,2) || seekdie();
    $dbfile->seek(0,0) || seekdie();
    $dbfile->guess_format();	# BUG: KLUGE: to skip over first line if it identifies columns
				# TBD: should remove; is an artifact of API using filehandles instead of filenames as interface


    #################################################################
    # the actual scan loop - code "borrowed" from scan_select. TBD: share

    my($InputTupleRef);
    local $db_algorithms::evaluate_table_and_field_specifier # NB: dynamic scoping via local, so that outside lambda function can call it
	= sub { 
	    my $arg = shift;
	    if( @_ ) { die "too many arguments / bad expression / for a single table join (no correlation names)\n"; }
	    my $val = $$InputTupleRef{$arg}; # TBD: do validity checking
	    if( !defined $val ) { return undef; }
	    return $val;
	};
    
    my %aggregate;
    $aggregate{'COUNT(*)'} = 0;


    # Loop over all aggregates, initializing
    foreach my $ag ( @$aggregate_list ) {
	die "need eval_name for aggregate" if not defined $ag->{eval_name};
	if( $ag->{eval_name} ne 'COUNT(*)' ) {
	    aggregate_initialize( aggregate_scratchpad => \%aggregate, aggregate_item => $ag );
	}
    }


    while( $InputTupleRef = $dbfile->get_next_tuple() ) {
	#disable diagnostics;
	my $wc = (ref  $parsed_where  eq 'CODE' ? &{$parsed_where}() : eval  $parsed_where  );
	#enable diagnostics;
	if( $wc ) {
	    # TBD: currently, only aggregate is 'COUNT(*)'
	    $aggregate{'COUNT(*)'}++;

	    # Loop over all aggregates, calculating arguments for this row
	    foreach my $ag ( @$aggregate_list ) {
		die "need eval_name for aggregate" if not defined $ag->{eval_name};
		if( $ag->{eval_name} ne 'COUNT(*)' ) {
		    my @vals = &eval_functions::eval_LIST(%{$ag->{argument_list}});
		    Carp::confess "aggregate function argument must evaluate to a scalar\n" if $#vals > 0;
		    aggregate_update( aggregate_scratchpad => \%aggregate, aggregate_item => $ag, value => $vals[0] );
		}
	    }


	}
    }


    # END OF the actual scan loop - code "borrowed" from scan_select. TBD: share
    #################################################################

    # save aggregates so that they can be dereferenced in actual scan
    $dbfile->{aggregate} = \%aggregate;

    #print "debug in calculate_aggregates $aggregate{'COUNT(*)'}:\n"
    #	  . Data::Dumper->Dump([\%aggregate],[qw(\%aggregate)]);

    #################################################################

    # Rewind dbfile before returning to user...
    $dbfile->seek(0,0) || seekdie();
    $dbfile->guess_format();	# BUG: KLUGE: to skip over first line if it identifies columns
    return;

}

sub aggregate_initialize{
    aggregate_operation( initialize => 1, @_);
}
sub aggregate_update{
    aggregate_operation( update => 1, @_);
}


sub is_aggregate_function_name {
    my($name) = @_;
    if( not defined $name ) { die "is_aggregate_function_name needs a name\n"; }

    return aggregate_operation( query => 1, aggregate_item => {eval_name=>$name} );
}

sub aggregate_operation{
    my %arg = @_;

    my $ag = $arg{aggregate_item};
    my $todate = $arg{aggregate_scratchpad};

    if( 0 ) {
	# skip
    }
    elsif( $ag->{eval_name} =~ m/^SUM\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{$ag->{eval_name}} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{$ag->{eval_name}} += $arg{value};
	    }
	    else {
		$todate->{$ag->{eval_name}} = undef; # propagate nulls
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^MAX\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{$ag->{eval_name}} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		if( !defined $todate->{"$ag->{eval_name} MAX-INITTED"} ) {
		    $todate->{"$ag->{eval_name} MAX-INITTED"} = 1;
		    $todate->{$ag->{eval_name}} = $arg{value};
		}
		elsif( $arg{value} > $todate->{$ag->{eval_name}} ) {
		    $todate->{$ag->{eval_name}} = $arg{value};
		}
	    }
	    else {
		$todate->{$ag->{eval_name}} = undef; # propagate nulls
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^MIN\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{$ag->{eval_name}} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		if( !defined $todate->{"$ag->{eval_name} MIN-INITTED"} ) {
		    $todate->{"$ag->{eval_name} MIN-INITTED"} = 1;
		    $todate->{$ag->{eval_name}} = $arg{value};
		}
		elsif( $arg{value} < $todate->{$ag->{eval_name}} ) {
		    $todate->{$ag->{eval_name}} = $arg{value};
		}
	    }
	    else {
		$todate->{$ag->{eval_name}} = undef; # propagate nulls
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^COUNT\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{$ag->{eval_name}} = 0;
        }
	elsif( $arg{update} ) {
	    if( defined $arg{value} ) {
		$todate->{$ag->{eval_name}}++;
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^AVG\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{"$ag->{eval_name} AVG_SUM"} = 0;
	    $todate->{"$ag->{eval_name} AVG_COUNT"} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{"$ag->{eval_name} AVG_SUM"} += $arg{value};
		$todate->{"$ag->{eval_name} AVG_COUNT"}++;
		$todate->{"$ag->{eval_name}"} = $todate->{"$ag->{eval_name} AVG_SUM"} / $todate->{"$ag->{eval_name} AVG_COUNT"};
	    }
	    else {
		$todate->{"$ag->{eval_name} AVG_SUM"} = undef;
		$todate->{"$ag->{eval_name} AVG_COUNT"} = undef;
		$todate->{"$ag->{eval_name}"} = undef;
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^ARITHMETIC_MEAN\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{"$ag->{eval_name} ARITHMETIC_MEAN_SUM"} = 0;
	    $todate->{"$ag->{eval_name} ARITHMETIC_MEAN_COUNT"} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{"$ag->{eval_name} ARITHMETIC_MEAN_SUM"} += $arg{value};
		$todate->{"$ag->{eval_name} ARITHMETIC_MEAN_COUNT"}++;
		$todate->{"$ag->{eval_name}"} = $todate->{"$ag->{eval_name} ARITHMETIC_MEAN_SUM"} / $todate->{"$ag->{eval_name} ARITHMETIC_MEAN_COUNT"};
	    }
	    else {
		$todate->{"$ag->{eval_name} ARITHMETIC_MEAN_SUM"} = undef;
		$todate->{"$ag->{eval_name} ARITHMETIC_MEAN_COUNT"} = undef;
		$todate->{"$ag->{eval_name}"} = undef;
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^HARMONIC_MEAN\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{"$ag->{eval_name} HARMONIC_MEAN_SUM"} = 0;
	    $todate->{"$ag->{eval_name} HARMONIC_MEAN_COUNT"} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{"$ag->{eval_name} HARMONIC_MEAN_SUM"} += (1/$arg{value});
		$todate->{"$ag->{eval_name} HARMONIC_MEAN_COUNT"}++;
		$todate->{"$ag->{eval_name}"} = 1/($todate->{"$ag->{eval_name} HARMONIC_MEAN_SUM"} / $todate->{"$ag->{eval_name} HARMONIC_MEAN_COUNT"});
	    }
	    else {
		$todate->{"$ag->{eval_name} HARMONIC_MEAN_SUM"} = undef;
		$todate->{"$ag->{eval_name} HARMONIC_MEAN_COUNT"} = undef;
		$todate->{"$ag->{eval_name}"} = undef;
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^GEOMETRIC_MEAN\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{"$ag->{eval_name} GEOMETRIC_MEAN_LOG_SUM"} = 0;
	    $todate->{"$ag->{eval_name} GEOMETRIC_MEAN_COUNT"} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{"$ag->{eval_name} GEOMETRIC_MEAN_LOG_SUM"} += log($arg{value});
		$todate->{"$ag->{eval_name} GEOMETRIC_MEAN_COUNT"}++;
		$todate->{"$ag->{eval_name}"} = exp($todate->{"$ag->{eval_name} GEOMETRIC_MEAN_LOG_SUM"} / $todate->{"$ag->{eval_name} GEOMETRIC_MEAN_COUNT"});
	    }
	    else {
		$todate->{"$ag->{eval_name} GEOMETRIC_MEAN_LOG_SUM"} = undef;
		$todate->{"$ag->{eval_name} GEOMETRIC_MEAN_COUNT"} = undef;
		$todate->{"$ag->{eval_name}"} = undef;
	    }
	}
    }
    elsif( $ag->{eval_name} =~ m/^STDDEV\(.*\)/i ) {
	if( $arg{query} ) {
	    return 1;
	}
	elsif( $arg{initialize} ) {
	    $todate->{"$ag->{eval_name} STDDEV_SUM_SQ"} = 0;
	    $todate->{"$ag->{eval_name} STDDEV_SUM"} = 0;
	    $todate->{"$ag->{eval_name} STDDEV_COUNT"} = 0;
        }
	elsif( $arg{update} ) {
	    if( exists($todate->{$ag->{eval_name}}) && !defined($todate->{$ag->{eval_name}}) ) {
		$todate->{$ag->{eval_name}} = undef;
	    }
	    elsif( defined $arg{value} ) {
		$todate->{"$ag->{eval_name} STDDEV_SUM_SQ"} += $arg{value}**2;
		$todate->{"$ag->{eval_name} STDDEV_SUM"} += $arg{value};
		$todate->{"$ag->{eval_name} STDDEV_COUNT"}++;
		$todate->{"$ag->{eval_name}"} = sqrt(
		    (($todate->{"$ag->{eval_name} STDDEV_SUM_SQ"} * $todate->{"$ag->{eval_name} STDDEV_COUNT"})
		    -($todate->{"$ag->{eval_name} STDDEV_SUM"}**2))
		    /($todate->{"$ag->{eval_name} STDDEV_COUNT"}**2)
                    );
			
	    }
	    else {
		$todate->{"$ag->{eval_name} STDDEV_SUM_SQ"} = undef;
		$todate->{"$ag->{eval_name} STDDEV_SUM"} = undef;
		$todate->{"$ag->{eval_name} STDDEV_COUNT"} = undef;
		$todate->{"$ag->{eval_name}"} = undef;
	    }
	}
    }
    else {
	if( $arg{query} ) {
	    return 0;
	}
	else {
	    die "unknown aggregate $ag->{eval_name}\n";
	}
    }
}




    

"KLUGE: return truth from module";




