#$Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/eval_functions.pm,v 1.2 1998/05/12 19:35:50 glew Exp $;

package eval_functions;


# Deduce location of libraries distributed with package
use FindBin;
BEGIN
{
    my $install_root = $FindBin::RealBin;
    $install_root =~ s:(.*/perl-sql)/.*:$1:;
    unshift (@INC, "$install_root/lib");
}

use Carp;
use Data::Dumper;
use strict;
use Perl_SQL::plain_text_database;
use Perl_SQL::db_algorithms;



# KLUGE:
# need to pass database file in to evaluation functions,
# which currently only have a node argument.
# Currently doing this by the following global.
# Note: NOT intended for use in db_algorithms package, 
# only for use in 

$eval_functions::kluge_database_file_being_processed = {};





#################################################################################################

# functions to evaluate SQL expression tree's with.
#
# each given node of the expression tree is a hash struct.
# each node hash struct contains an eval_function element.
# each eval_function is called with the node itself as argument.
#
# i.e.        &{noderef->eval_function}( %$noderef )
#


sub eval_NODE {
    my($node_ref) = @_;		# only exception to the rule?
    &{$node_ref->{eval_function}}(%$node_ref)
}

sub eval_ERROR {
    my %arg = @_;
    confess "Internal Error: error evaluating perl-SQL expression trees\n"
	. "(probably corresponds to unimplemented syntax - in grammar, but not coded yet)\n"
	    . Data::Dumper->Dump([\%arg],['expression_node']);
}




sub eval_NUMBER {
    my %arg = @_;
    if( !exists $arg{string} ) { die "eval_NUMBER no string\n"; }
    if( $arg{subtype} ne 'decimal integer' 
       )  { die "eval_NUMBER only works for perl integers. Nothing fancy yet. TBD: fix\n"; }
    return 0+$arg{string};	# TBD: works for decimal and FP numbers, not for funny bases. 
}


sub eval_STRING {
    my %arg = @_;
    return eval $arg{string};	# TBD: needs to handle '' -> \', etc. syntaxes
}




#
# TBD: evals
#
# we are doing far too many evals
# - this will give us slow performance.
# we would probably be better off making
# at least some expressions into strings that
# can be directly eval'ed.
#
# however, what we have suffices for prototyping.
#

sub eval_null_propagating_perl_prefix_operator {
    my %arg = @_;
    my $operand = eval_NODE($arg{operand});
    if( !defined $operand ) { return undef; }

    my $operator = $arg{perl_operator} || $arg{string} || die "unknown operator";

    return (eval "$operator \$operand")||0;
}
sub eval_null_propagating_perl_infix_operator {
    my %arg = @_;
    my $left = eval_NODE($arg{left_operand});
    my $right = eval_NODE($arg{right_operand});
    if( !defined $left ) { return undef; }
    if( !defined $right ) { return undef; }

    my $operator = $arg{perl_operator} || $arg{string} || die "unknown operator";

    return (eval "\$left $operator \$right")||0;
}
sub eval_standard_perl_infix_operator {
    my %arg = @_;
    my $left = eval_NODE($arg{left_operand});
    my $right = eval_NODE($arg{right_operand});

    my $operator = $arg{perl_operator} || $arg{string} || die "unknown operator";

    return (eval "\$left $operator \$right")||0;
}


sub eval_NAME {
    my %arg = @_;

    if( exists $arg{eval_name} ) {
	#print "debug: eval_NAME $arg{eval_name}\n";
	return $arg{eval_name};
    }
    else {
      Carp::confess "Internal error: eval_NAME on a node that has no eval_name (lvalue)\n"
	  . Data::Dumper->Dump([\%arg],['eval_NAME']);
	
    }
}



sub eval_SELECT {
    my %arg = @_;

    my @aggregates_found = ();
    my $selection_all_aggregates = 0;

    #print "debug SELECT\n"
    #	  . Data::Dumper->Dump([\%arg],['eval_SELECT-node']);

    if( !exists $arg{select_operator} 
       || !exists $arg{select_operator}->{operator}
       || $arg{select_operator}->{operator} ne 'SELECT' ) 
    {
      Carp::confess "Internal Error: eval_SELECT with invalid SELECT operator $arg{select_operator}->{operator}\n"
	  . Data::Dumper->Dump([\%arg],['eval_SELECT-node']);
    }
    if( exists $arg{select_operator}->{quantifier} ) {
	die "Unimplemented feature (in gramar, not coded) SELECT quantifier (e.g. SELECT ALL or SELECT DISTINCT)\n";
    }

    ##############################################################
    # handle selection - variables, expressions, to be output

    my $selection;
    my @selection_fieldnames_order = ();
    if( !exists $arg{selection} ) {
	$selection = '*';
    }
    elsif( exists $arg{selection}->{eval_name} &&  $arg{selection}->{eval_name} eq "WILDCARD('*')" ) {
	$selection = '*';
    }
    elsif( exists $arg{selection}->{eval_name} &&  $arg{selection}->{eval_name} =~ m/WILDCARD\(.*\)/ ) {
	die "Unsupported wildcard $arg{selection}->{eval_name}\n";
    }
    else {
      Carp::confess "Internal error: Missing as_list" if !exists $arg{selection}->{as_list};


	#TBD # if we have a defined selection, use CSV as default
	#TBD # (rationale: less typing for gnuplot usage model)
	#TBD $::PTD_OUT_FORMAT = &plain_text_database::string_to_database_format("csv");
	
	$selection = {};
	my $asp;

	# Validate selection - no multiples, create default AS name if none exists
	
	foreach $asp ( @{$arg{selection}->{as_list}} ) {
	  Carp::confess "Internal error: Missing as_name" if !exists $asp->{as_name};
	  Carp::confess "Internal error: Missing as_value" if !exists $asp->{as_value};
	  Carp::confess "Internal error: Missing eval_function" if !exists $asp->{as_value}->{eval_function};
	    
	    die "Multiply defined fieldname '$asp->{as_name}' in selection" if exists $selection->{$asp->{as_name}};
	    
	    die "Multiply defined fieldname '$asp->{as_name}' in selection" if exists $selection->{$asp->{as_name}};
	    
	    $selection->{$asp->{as_name}} = sub { return eval_NODE($asp->{as_value}) };  

	    push @selection_fieldnames_order, $asp->{as_name};
	}

	# Determine if there are any aggregate functions in the selection
	# and prepare to process them

	$selection_all_aggregates = 1;
	foreach $asp ( @{$arg{selection}->{as_list}} ) {
	    push @aggregates_found, detect_aggregates( $asp );
	    if( not $asp->{as_value}->{is_aggregate} ) { $selection_all_aggregates = 0; }
	}
	
	# The actual aggregate scan work is done elsewhere, just before scan



    }
    #print Data::Dumper->Dump([\$selection],['debug-selection']);



    ##############################################################
    # WHERE clause
    
    my $where;
    if( defined $arg{WHERE} ) {
	die "Internal Error: WHERE missing predicate" if !defined $arg{WHERE}->{predicate};
	$where = sub { return eval_NODE($arg{WHERE}->{predicate});};
    }
    else {
	$where = '1'; 
    }

    ##############################################################
    # INTO clause

    my $into;
    if( defined $arg{INTO} ) {
	die "Internal Error: INTO missing target" if !defined $arg{INTO}->{target};
	$into = eval_NAME(%{$arg{INTO}->{target}});
    }
    else {
	$into = '-';		# TBD: STDIN
    }
    my($ptd_out) = new plain_text_database ">$into";
    die "error opening $into\n" if not $ptd_out;

    my %dbformat = ();
    if( defined($::PTD_OUT_FORMAT) ) { %dbformat = %$::PTD_OUT_FORMAT };
    if( !defined $dbformat{FORMAT} ) { $dbformat{FORMAT} = "multiline_named_tuples" };
    if( defined @selection_fieldnames_order ) { $dbformat{FIELD_NAMES} = \@selection_fieldnames_order };
    $ptd_out->database_format( %dbformat );
    if( $::print_format_headers_for_database ) {
	$ptd_out->print_format();
    }


    ##############################################################
    # FROM clause - distinguishing simple FROM from JOIN

    my $from; 
    my $join;

    if( defined $arg{FROM} ) {

        Carp::confess "Internal Error: FROM missing source" if !defined $arg{FROM}->{source};
        Carp::confess "Internal Error: FROM missing table_expr" if !defined $arg{FROM}->{source}->{table_expr};

	if( !defined $arg{FROM}->{source}->{table_expr}->{join_op} ) {
	    $join = 0;
	
	    die "correlation names not yet supported for single table joins\n" if exists $arg{FROM}->{source}->{correlation_name};
	    # TBD: save the correlation name for something?
	    $from = eval_NAME(%{$arg{FROM}->{source}->{table_expr}});
	    die "Error: file/table 'FROM $from' doesn't exist\n" if ! -r $from;
	}
	else {
	    $join = 1;
	}
      }
    else {
	$from = '-';		# TBD: STDIN
	$join = 0;
    }

    die "FROM = INTO = $from not currently supported\n"
	if $join == 0 and $from eq $into and $from ne '-';



    ##########################################################
    # TBD: error if any clauses left unhandled


    die "error: limitation: tbd: aggregate functions only supported for simple scans\n"
	if( @aggregates_found && $join );



    ###############################################################
    # Now do either the simple SELECT scan setup, or the SELECT/JOIN


    if( $join == 0 ) {

	##########################################################
	# Ordinary scan select
	
	my($ptd) = new plain_text_database "<$from";
	die "error opening $from\n" if not $ptd;
	$ptd->guess_format();
	
	$eval_functions::kluge_database_file_being_scanned = $ptd; # KLUGE

	if( @aggregates_found ) {
	  db_algorithms::calculate_aggregates( dbfile => $ptd, aggregates => \@aggregates_found, where => $where );
	}
    
        db_algorithms::scan_select( database => $ptd, 
				   output => $ptd_out,
				   where => $where, 
				   'select' => $selection,	# TBD: use selection
				   selection_all_aggregates => $selection_all_aggregates,
				   );
       $eval_functions::kluge_database_file_being_scanned = {}; # KLUGE
    } 
    else {

	##########################################################
	# JOINs

	my $join_expr = $arg{FROM}->{source}->{table_expr};

	# Setting up the left and right tables

	die "Limitation: JOIN only supports simple tables. left not simple\n"
	    if( ! $join_expr->{left}->{table_expr}->{type}->{simple_identifier} );
	die "Limitation: JOIN only supports simple tables. right not simple\n"
	    if( ! $join_expr->{right}->{table_expr}->{type}->{simple_identifier} );

	my $left_table = eval_NAME(%{$join_expr->{left}->{table_expr}});
	my $right_table = eval_NAME(%{$join_expr->{right}->{table_expr}});

	my $left_table_correlation_name = $join_expr->{left}->{correlation_name};
	my $right_table_correlation_name = $join_expr->{right}->{correlation_name};


	my($ptd1) = new plain_text_database "<$left_table";
	die "error opening $left_table\n" if not $ptd1;
	$ptd1->guess_format();

	my($ptd2) = new plain_text_database "<$right_table";
	die "error opening $right_table\n" if not $ptd2;
	$ptd2->guess_format();

	my %correlation_names = (
				 PRIMARY_left_table_correlation_name=>($left_table_correlation_name||$left_table),
				 PRIMARY_right_table_correlation_name=>($right_table_correlation_name||$right_table),
				 ($left_table_correlation_name ? ($left_table_correlation_name=> -1) : () ),
				 ($right_table_correlation_name ? ($right_table_correlation_name=> +1) : () ),
				 $left_table=>-1, 
				 $right_table=>+1,
				 );

	# Setting up the JOIN ON clause

	my $on = $join_expr->{on} || die "Internal error: missing ON clause\n";
	my $on_as_name = $on->{as_mapping}->{as_name} || die "Internal error: ON/AS name\n";
	my $on_expr = $on->{as_mapping}->{as_value} || die "Internal error: ON/AS value\n";

	my $on_op = $on_expr->{perl_operator};
	if( $on_op eq 'eq' ) {
	    # placeholder
	}
	elsif( $on_op eq '==' ) {
	    # placeholder
	}
	# TBD: other equality operators (special null handling, etc?)
	else {
	    die "error: JOIN only supports Perl equality operators for equijoins (eq, ==) and SQL-like = => eq, not $on_op\n";
	}
	
	# Looking at the JOIN ON expression,
	# determining which variable comes from which table

	my $on_left = $on_expr->{left_operand};
	my $on_right = $on_expr->{right_operand};

	# TBD: desirable enhancement: permit expressions to be used to determine
	# the names of the variables to be joined on, using a COLUMN() case like TABLE()

	die "error: JOIN only accepts column = column (left not)\n" if !$on_expr->{left_operand}->{is_column_ref};
	die "error: JOIN only accepts column = column (right not)\n" if !$on_expr->{right_operand}->{is_column_ref};

	my $lcc = $on_expr->{left_operand}->{column_ref_components};
	my $rcc = $on_expr->{right_operand}->{column_ref_components};

	die "error: JOIN requires table.column operands in JOIN ON clause (left not)\n"
	    if $#$lcc != 1;
	die "error: JOIN requires table.column operands in JOIN ON clause (right not)\n"
	    if $#$rcc != 1;
	
	my $olt = $$lcc[0];
        my $ort = $$rcc[0];

        my $join_on_var_left;
	my $join_on_var_right;

	if( !defined $correlation_names{$olt} ) {
	    die "error: attempting to join on non-table $olt\n";
	}
	elsif( !defined $correlation_names{$ort} ) {
	    die "error: attempting to join on non-table $ort\n";
	}
	elsif( $correlation_names{$olt} < 0 && $correlation_names{$ort} > 0 ) {
	    $join_on_var_left = $$lcc[1];
	    $join_on_var_right = $$rcc[1];
	}
	elsif( $correlation_names{$olt} > 0 && $correlation_names{$ort} < 0 ) {
	    $join_on_var_left = $$rcc[1];
	    $join_on_var_right = $$lcc[1];
	}
	else {
	    die "error: need to join on a variable from each table\n";
	}

    

	#
	# JOIN: setup finished, actually do it
	# 
    
	# TDB: select the whole thing
	db_algorithms::sorted_merge_join(
			   'join' => $join_expr->{join_op},
			   left =>$ptd1,
			   right =>$ptd2,
			   join_on_left_var => $join_on_var_left,
			   join_on_right_var => $join_on_var_right,
			   join_on_var_label => $on_as_name,
			   join_cmp => $on_op,
					 
			   output => $ptd_out,
			   where => $where,
			   'select' => $selection,

			   # TBD: my handling of correlation names is redundant
			   # but distinguishes between many to one for specification, and 1to1 for output default
			   correlation_names=>\%correlation_names,
			   );

    }
}





# TBD: share the common stuff between eval_SELECT, eval_DELETE, eval_UPDATE, eval_INSERT, and eval_ALTER

sub eval_DELETE {
    my %arg = @_;

    #print "debug DELETE\n"
    #      . Data::Dumper->Dump([\%arg],['eval_DELETE-node']);

    # TBD: typecheck?  add type fields to node, or rely on eval_function == eval_DELETE?

    my $where_not_to_delete;
    if( defined $arg{WHERE} ) {
	die "Internal Error: WHERE missing predicate" if !defined $arg{WHERE}->{predicate};
	$where_not_to_delete = sub { return ! eval_NODE($arg{WHERE}->{predicate});}; 
    }
    else {
	$where_not_to_delete = '0'; 
    }

    my $from; 
    if( defined $arg{FROM} ) {
	die "correlation names not yet supported for single table joins\n" if exists $arg{FROM}->{source}->{correlation_name};
	# TBD: save the correlation name for something?
	$from = eval_NAME(%{$arg{FROM}->{source}->{table_expr}});
	die "Error: file/table 'FROM $from' doesn't exist\n" if ! -r $from;
    }
    else {
	$from = '-';		# TBD: STDIN
    }
    my($ptd) = new plain_text_database "<$from";
    die "error opening $from\n" if not $ptd;
    $ptd->guess_format();

    my $into;
    if( defined $arg{INTO} ) {
	die "Internal Error: INTO missing target" if !defined $arg{INTO}->{target};
	$into = eval_NAME(%{$arg{INTO}->{target}});
    }
    else {
	$into = '-';		# TBD: STDIN
    }
    my($ptd_out) = new plain_text_database ">$into";
    die "error opening $into\n" if not $ptd_out;
    $ptd_out->database_format( %$ptd ); # TBD: specify format somewhere
    if( $::print_format_headers_for_database ) {
	$ptd_out->print_format();
    }

    die "FROM = INTO = $from not currently supported\n"
	if $from eq $into and $from ne '-';

    db_algorithms::scan_select( database => $ptd, 
			       output => $ptd_out,
			       where => $where_not_to_delete, 
			       'select' => '*',	
			       );
}

sub eval_column_ref {
    my %arg = @_;
    
    if( !defined $arg{column_ref_components} ) {
      Carp::confess "Internal error: eval_column_ref without column_ref_components\n"
	  . Data::Dumper->Dump([\%arg],['eval_column_ref']);
    }

    # CAREFUL!!! KLUGE:
    # Sensitive, unusual (i.e. probably bad) code:
    # Each db_algorithm defines its own method of accessing
    # a column_ref (a specifier like "column", or "table.column", etc.).
    # The original code used eval of string, and placed this method
    # in a variable $evaluate_table_and_field_specifier.
    # This was originally lexically scoped, via "my",
    # but to make it work when the expression evaluator wasn't lexically nested,
    # but was initiated via an external lambda function that walked
    # an expression tree,
    # $evaluate_table_and_field_specifier was made into a dynamically scoped (local) variable.
    # TBD: clean this up!!!

    #BROKEN: no strict 'refs';
    #BROKEN: my $retval = &${'evaluate_table_and_field_specifier'}( $arg{column_ref_components} );
    #TBD: wanted to avoid specifying a package name, so could use true dynamic scoping.
    # However, I did not want to give up struct refs to do so.
    # TBD: may have to parametrize closure, making it a bit less general.
    #BROKEN: strict 'refs';

    my $retval = &$db_algorithms::evaluate_table_and_field_specifier( @{$arg{column_ref_components}} );


    #print "DEBUG in eval_column_ref\n"
    #	  . Data::Dumper->Dump([\%arg],['eval_column_ref'])
    #	  . Data::Dumper->Dump([$retval],['retval'])
    #	  . Data::Dumper->Dump([$db_algorithms::evaluate_table_and_field_specifier],['$db_algorithms::evaluate_table_and_field_specifier']);


    return $retval;  

    # TBD: handle both lvalue and rvalue versions of this.
}




sub eval_FUNCTION {
    my %arg = @_;

    my $function_name = $arg{function_name}->{eval_name} || die "internal error: missing function_name\n";
    my $argument_list = $arg{argument_list}->{list} || die "internal error: missing argument_list\n";

    my @arglist = ();
    foreach my $a ( @$argument_list ) {
	my $ar = eval_NODE($a);
	if( not defined $ar ) { $ar = "undef" }
	push @arglist, $ar;
    }
    
    
    my $result = eval "$function_name(" . (defined @arglist ? join(',',@arglist) : '') . ")";
    
    return $result;
}

sub eval_COUNT_STAR {
    my %arg = @_;

    # TBD: as in the other eval functions, I really need to have a database argument passed in
    # Currently doing it via a global

    Carp::confess "no aggregates defined\n" 
	if not defined $eval_functions::kluge_database_file_being_scanned->{aggregate};
    Carp::confess "no COUNT(*) aggregate defined\n" 
	if not defined $eval_functions::kluge_database_file_being_scanned->{aggregate}->{'COUNT(*)'};

    return $eval_functions::kluge_database_file_being_scanned->{aggregate}->{'COUNT(*)'};
}


sub eval_AGGREGATE_FUNCTION {
    my %arg = @_;

    # TBD: as in the other eval functions, I really need to have a database argument passed in
    # Currently doing it via a global

    Carp::confess "no aggregates defined\n" 
	if not defined $eval_functions::kluge_database_file_being_scanned->{aggregate};
    Carp::confess "need eval_name" 
	if !defined $arg{eval_name};
    Carp::confess "no $arg{eval_name} aggregate defined\n" 
	if not defined $eval_functions::kluge_database_file_being_scanned->{aggregate}->{$arg{eval_name}};

    return $eval_functions::kluge_database_file_being_scanned->{aggregate}->{$arg{eval_name}};
}


sub is_aggregate_function_name {
    return db_algorithms::is_aggregate_function_name( @_ );
}



# walk the expression tree, looking for aggregate functions
# returns a list of aggregate functions that need to be evaluated
sub detect_aggregates {
    my ($root) = @_;
    my @retlist = ();

    if( ref($root) eq 'HASH' ) {
	if( exists $root->{is_aggregate} ) {
	    if( $root->{is_aggregate} == 1 ) {
		# Definitely an aggregate - marked as suc
		push @retlist, $root;
	    }
	    else {
		# Possibly an aggregate
		# - query 
		die "error: tbd: curently not implementing possible aggregate query functions\n";
		#push @retlist, $root;
	    }
	}
	my @subretlist = ();
	foreach my $k ( keys %$root ) {
	    @subretlist = detect_aggregates($root->{$k});
	    die "error: tbd: Currently do not support nested aggregate functions\n" if @retlist and @subretlist;
	    push @retlist, @subretlist;
	}
	return @retlist;
    }
    elsif( ref($root) eq 'ARRAY' ) {
	foreach my $a ( @$root ) {
	    push @retlist, detect_aggregates($a);
	}
	return @retlist;
    }
    else {
	return ();
    }
}


# evaluate each element of a list
# (typically a comma separated list)
# and return a Perl list of the results.
#
# not quite proper, but a hack and a convenience

sub eval_LIST {
    my %arg = @_;
    
    my @retval = ();

    die "eval_LIST needs a list\n" if not defined $arg{list};
    foreach my $l ( @{$arg{list}} ) {
	push @retval, eval_NODE($l);
    }
    return @retval;
}





"Kluge return value from module";


