#!/s/perl-5.004/bin/perl -w

## $Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/nice_IO.pm,v 1.2 1998/05/12 19:35:51 glew Exp $

##
## nice_IO
## 
## a wrapper around (a very limited number of APIs of)
## standard I/O, providing abilities such as
## (1) automatic deletion of comments
## (1) automatic deletion of blank lines
## may be extended, in the future, to include
## (?) line-number counting (for nice error messages)
## (?) multiple lines of scan ahead and push pack
##
## TBD: make full-fledged, inheriting from, and superceding, IO:File
##


package nice_IO;

use IO::File;
@ISA = qw(IO::File);

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};

    bless ($self, $class);

    $self->{NICE_IO_FILEHANDLE} = new IO::File;
    if( !defined $self->{NICE_IO_FILEHANDLE} ) {
	return undef;
    }

    $self->{NICE_IO}   = 1;

    $self->filter("default");

    $self->{ungetlines} = [];

    if( defined @_ ) {
	$self->open(@_) || return undef;
    }

    return $self;
}

sub open {
    my($self) = shift @_;

    $self->{NICE_IO_FILENAME} = $_[0];
    return IO::File::open $self->{NICE_IO_FILEHANDLE},@_;
}


sub seek {
    my $self = shift;
    my $offset = shift;
    my $whence = shift;
    die "Too many arguments to nice_IO::seek\n" if @_;
    
    $self->{ungetlines} = [];
    return IO::Seekable::seek($self->{NICE_IO_FILEHANDLE},$offset,$whence);
}



# TBD: make an object?
sub getline {
    my($self) = @_;

    for(;;) {
	my($line);

	# get the next line
	if( defined( $line = shift @{$self->{ungetlines}} ) ) {
	    # ungotten 
	}
	elsif( defined($line = $self->{NICE_IO_FILEHANDLE}->getline) ) {
	    # actually read in from file;	# TBD: change file
	}
	else {
	    # at end of both ungotten and file
	    return undef;
	}


	# if raw mode, return immediately
	if( ! $self->filter() ) {
	    return $line;
	}
	
	# cook the data:

	if( $self->{'chomp'} ) {
	    chomp $line;
	}

	if( $self->{comment} ) {
	    # Remove comments
	    # TBD: scan some comments
	    # and/or save them to output
	    $line =~ s/$self->{comment}.*//;
	}

	if( $self->remove_whitespace ) {
	    # Remove whitespace
	    # TBD: parametrize - in some formats, don't want to
	    $line =~ s/^[ \t]+//;
	    $line =~ s/[ \t]+$//;
	}
	
	if( $line ne "" ) { return $line; }
    }
}

# ungetlines:
# method to read lines that have been "pushed back"
# and/or to actually push back, ahead or behind of already
# pushed back data.
# Usage: $fh->ungetlines( "head-data", $fh->ungetlines(), "tail-data" );
# Note: you may or may not need to worry about newlines, depending on consumer.
sub ungetlines {
    my $self = shift;

    my(@list) = @_;

    if (@_) { $self->{ungetlines} = \@list; }

    return 1;
}

sub getlines {
    my($self) = @_;
    my($line);
    my(@result) = ();

    wantarray or
	warn 'Cannot call nice_IO::getlines in a scalar context, use Perl_SQL::nice_IO::getline';


    while( defined($line = $self->getline) ) {
	@result = (@result, $line);
    }
    return @result;
}


# stubs to pass other methods to IO::File
# TBD: figure out how to inheritance automatic,
# in this case when IO::File handles are globs, not hashes.
# ??: use GLOB (sym), and then have a separate table that 
# maps glob to assisting data structure?
sub close {
    my $self = shift @_;

    $self->{NICE_IO_FILEHANDLE}->close(@_);
}
sub print {
    my $self = shift;

    $self->{NICE_IO_FILEHANDLE}->print(@_);
}
sub printf {
    my $self = shift;

    $self->{NICE_IO_FILEHANDLE}->printf(@_);
}



# per-instance control variables
sub comment {
    my $self = shift;
    my $old = $self->{comment};
    if (@_) { $self->{comment} = shift; }
    return $old;
}
sub remove_whitespace {
    my $self = shift;
    my $old = $self->{remove_whitespace};
    if (@_) { $self->{remove_whitespace} = shift; }
    return $old;
}
sub chomp {
    my $self = shift;
    my $old = $self->{'chomp'};
    if (@_) { $self->{'chomp'} = shift; }
    return $old;
}

# filter - enable filtering of I/O
# undef or 0 => no filtering, raw mode
# nonzero => use filtering, cooked mode
# 'default' => set to default filtering
sub filter {
    my $self = shift;
    my $old = $self->{filter};
    if (@_) { 
	$self->{'filter'} = shift;
	if($self->{'filter'} eq 'default') {
	    # default filtering should be placed somewhere
	    $self->{filter} = 1;
	    $self->comment('#');
	    $self->chomp(1);
	    $self->remove_whitespace(1);
	}
    }
    return $old;
}





########################################################################


    
"KLUGE: packages musrt return success";

