# $Header: /u/g/l/glew/cvsroot/perl-sql/lib/Perl_SQL/singleline_delimited_records.pm,v 1.1.1.1 1998/05/05 23:52:23 glew Exp $


# singleline_delimited_records.pm
#
# handle records of the form
# val1,val2,val3...
# e.g. comma-separated-variables, tab-separated-variables, etc.
#
# TBD: should I convert "undef" or "null" on input into a Perl undef?
# Currently handling nulls for output, but not on input => assymmetric.
# Unsure of right thing.
#


# Deduce location of libraries distributed with package
use FindBin;
BEGIN
{
    my $install_root = $FindBin::RealBin;
    $install_root =~ s:(.*/perl-sql)/.*:$1:;
    unshift (@INC, "$install_root/lib");
}


use Perl_SQL::nice_IO;
use Carp;
use Data::Dumper;
use strict;
use Perl_SQL::misc;

package singleline_delimited_records;

sub read_record {
    my($dbfile) = @_;
    my($file) = $dbfile->{NICE_IO_FILEHANDLE};

    return primitive_read_record( $file, 
				 delimiter_regexp => $dbfile->{FIELD_DELIMITER_REGEXP}, 
				 field_names => $dbfile->{FIELD_NAMES}
				 );
}

# primitive, takes a nice_IO filehandle
# and an option string, rather than a database handle
# not intended for use, only so that I can test
# without having all working
# TBD: the primitives probably should have been implemented
# as a base class and inherited from
sub primitive_read_record {
    my $file = shift @_;
    my %options = @_;

    my($line);

    $line = $file->getline();
    if( !defined $line ) { return undef };

    my $regexp = $options{delimiter_regexp} || '\s+';
    my(@fields) = split(/$regexp/, $line );

    my(@field_names) = ();
    if( defined $options{field_names} ) {
	@field_names = @{$options{field_names}};
    }

    my(%tuple) = ();
    
    for(my($i)=1;;$i++) {
	my($fn);
	my($fv);

	$fn = shift @field_names;
	$fv = shift @fields;
	if( !defined($fv) ) { last; } 

	if( !defined($fn) ) { 
	    $fn = "field" . $i; 
	    # remember field definitions
	    # TBD: this is probably not the right place to do this
	    # since the "primitive" doesn't really know about the
	    # calling environment. Better to do it outside,
	    # after tuple is returned. 
	    push @{$options{field_names}}, $fn;

	}

	$tuple{$fn} = $fv;
    }

    return \%tuple;
}

sub print_record {
    my($dbfile,$record) = @_;
    my($file) = $dbfile->{NICE_IO_FILEHANDLE};

    primitive_print_record( $dbfile, $record, field_delimiter => $dbfile->{FIELD_DELIMITER} );
}


# primitive, takes a nice_IO filehandle
# and an option string, rather than a database handle
# not intended for use, only so that I can test
# without having all working
# TBD: the primitives probably should have been implemented
# as a base class and inherited from

sub primitive_print_record {
    my($dbfile) = shift;
    my($file) = $dbfile->{NICE_IO_FILEHANDLE};
    my $record = shift;
    my %options = @_;


    my($fn);

    my @fieldnames_print_order;

    if( !defined $dbfile->{FIELD_NAMES} ) {
	@fieldnames_print_order = sort keys %$record;
    }
    else {
	@fieldnames_print_order = @{$dbfile->{FIELD_NAMES}};
    }
    my @fieldnames_print_order_copy = @fieldnames_print_order;
    
    my %tuple_copy = %$record;

    for(;;) {
	my $k = shift @fieldnames_print_order_copy;
	last if !defined $k;
	my $val =  &null_handling::to_string($record->{$k});
	$file->print( $val );
	delete $tuple_copy{$k};
	$file->print( $options{field_delimiter} ) if( @fieldnames_print_order_copy );
    }

    if( keys %tuple_copy ) {
	die "Error: Singleline Delimited Record (SDR) with more fields than expected\n"
	    . "Expected fields: " . join(':', @fieldnames_print_order_copy) . "\n"
		. "Found extra fields: " . join('|', keys %tuple_copy) . "\n";
    }

    $file->print( "\n" );
}



"KLUGE: packages must return a true value";












